import csv
import json
import logging
import sys
import time
import traceback
import i18n
from typing import List
from typing import Optional

from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.custom_exceptions import TransformationFieldMappingError
from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.manual_fee_fines_mapper import (
    ManualFeeFinesMapper,
)
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class ManualFeeFinesTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        feefines_map: str
        migration_task_type: str
        files: List[FileDefinition]
        feefines_owner_map: Optional[str]
        feefines_type_map: Optional[str]
        service_point_map: Optional[str]

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.fees_fines

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        csv.register_dialect("tsv", delimiter="\t")

        super().__init__(library_config, task_configuration, use_logging)
        self.object_type_name = self.get_object_type().name
        self.task_configuration = task_configuration
        self.check_source_files(
            self.folder_structure.legacy_records_folder, self.task_configuration.files
        )
        self.total_records = 0
        self.feefines_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder / self.task_configuration.feefines_map
        )
        self.results_path = self.folder_structure.created_objects_path
        self.failed_files: List[str] = []

        self.folio_keys = []
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(
            self.feefines_map
        )

        self.mapper = ManualFeeFinesMapper(
            self.folio_client,
            self.library_configuration,
            self.task_configuration,
            self.feefines_map,
            feefines_owner_map=self.load_ref_data_mapping_file(
                "account.ownerId",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.feefines_owner_map,
                self.folio_keys,
                False,
            ),
            feefines_type_map=self.load_ref_data_mapping_file(
                "account.feeFineId",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.feefines_type_map,
                self.folio_keys,
                False,
            ),
            service_point_map=self.load_ref_data_mapping_file(
                "feefineaction.createdAt",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.service_point_map,
                self.folio_keys,
                False,
            ),
            ignore_legacy_identifier=True,
        )

    def do_work(self):
        logging.info("Getting started!")
        for file in self.task_configuration.files:
            logging.info("Processing %s", file)
            try:
                self.process_single_file(file)
            except Exception as ee:
                error_str = (
                    f"Processing of {file} failed:\n{ee}."
                    "Check source files for empty rows or missing reference data"
                )
                logging.exception(error_str)
                self.mapper.migration_report.add("FailedFiles", f"{file} - {ee}")
                sys.exit()

    def process_single_file(self, file_def: FileDefinition):
        full_path = self.folder_structure.legacy_records_folder / file_def.file_name
        with open(full_path, encoding="utf-8-sig") as records_file:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of files processed")
            )
            start = time.time()

            for idx, record in enumerate(self.mapper.get_objects(records_file, full_path)):
                try:
                    if idx == 0:
                        logging.info("First legacy record:")
                        logging.info(json.dumps(record, indent=4))

                    self.mapper.report_legacy_mapping_no_schema(record)

                    composite_feefine, legacy_id = self.mapper.do_map(
                        record, f"Row {idx + 1}", FOLIONamespaces.fees_fines
                    )

                    self.mapper.perform_additional_mapping(legacy_id, composite_feefine, record)

                    self.mapper.report_folio_mapping(
                        composite_feefine, self.mapper.composite_feefine_schema
                    )

                    self.mapper.store_objects(composite_feefine)

                    if idx == 0:
                        logging.info("First FOLIO record:")
                        logging.info(json.dumps(composite_feefine, indent=4))

                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as data_error:
                    self.mapper.handle_transformation_record_failed_error(idx, data_error)
                except TransformationFieldMappingError as mapping_error:
                    self.mapper.handle_transformation_field_mapping_error(idx, mapping_error)

                except AttributeError as attribute_error:
                    traceback.print_exc()
                    logging.fatal(attribute_error)
                    logging.info("Quitting...")
                    sys.exit(1)
                except Exception as exception:
                    self.mapper.handle_generic_exception(idx, exception)

                self.print_progress(idx, start)

    def wrap_up(self):
        logging.info("Done. Transformer wrapping up...")
        self.extradata_writer.flush()
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            logging.info(
                "Writing migration- and mapping report to %s",
                self.folder_structure.migration_reports_file,
            )
            self.mapper.migration_report.write_migration_report(
                i18n.t("Manual fee/fine transformation report"),
                migration_report_file,
                self.start_datetime,
            )

            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )

        self.clean_out_empty_logs()



import json
import logging
from typing import Set

import httpx
import i18n
from folio_uuid import FOLIONamespaces
from folioclient import FolioClient
from pymarc import Field
from pymarc import Record
from pymarc import Subfield

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.migration_report import MigrationReport


class HRIDHandler:
    def __init__(
        self,
        folio_client: FolioClient,
        handling: HridHandling,
        migration_report: MigrationReport,
        deactivate035_from001: bool,
    ):
        self.unique_001s: Set[str] = set()
        self.deactivate035_from001: bool = deactivate035_from001
        self.hrid_path = "/hrid-settings-storage/hrid-settings"
        self.folio_client: FolioClient = folio_client
        self.handling: HridHandling = handling
        self.migration_report: MigrationReport = migration_report
        self.hrid_settings = self.folio_client.folio_get_single_object(self.hrid_path)
        self.instance_hrid_prefix = self.hrid_settings["instances"].get("prefix", "")
        self.instance_hrid_counter = self.hrid_settings["instances"]["startNumber"]
        self.holdings_hrid_prefix = self.hrid_settings["holdings"].get("prefix", "")
        self.holdings_hrid_counter = self.hrid_settings["holdings"]["startNumber"]
        self.items_hrid_prefix = self.hrid_settings["items"].get("prefix", "")
        self.items_hrid_counter = self.hrid_settings["items"]["startNumber"]
        self.common_retain_leading_zeroes: bool = self.hrid_settings["commonRetainLeadingZeroes"]
        logging.info(f"HRID handling is set to: '{self.handling}'")

    def handle_hrid(
        self,
        namespace: FOLIONamespaces,
        folio_record: dict,
        marc_record: Record,
        legacy_ids: list[str],
    ) -> None:
        """Create HRID if not mapped. Add hrid as MARC record 001

        Args:
            namespace (FOLIONamespaces): determening the type of hrid setting to update
            folio_record (dict): _description_
            marc_record (Record): _description_
            legacy_ids (list[str]): _description_

        Raises:
            TransformationProcessError: _description_
        """
        if self.enumerate_hrid(marc_record):
            self.generate_enumerated_hrid(folio_record, marc_record, legacy_ids, namespace)
        elif self.handling == HridHandling.preserve001:
            self.preserve_001_as_hrid(folio_record, marc_record, legacy_ids, namespace)
        else:
            raise TransformationProcessError("", f"Unknown HRID handling: {self.handling}")

    def generate_enumerated_hrid(
        self,
        folio_record: dict,
        marc_record: Record,
        legacy_ids: list[str],
        namespace: FOLIONamespaces,
    ):
        folio_record["hrid"] = self.get_next_hrid(namespace)
        new_001 = Field(tag="001", data=folio_record["hrid"])
        self.handle_035_generation(
            marc_record, legacy_ids, self.migration_report, self.deactivate035_from001
        )
        marc_record.add_ordered_field(new_001)
        self.migration_report.add("HridHandling", i18n.t("Created HRID using default settings"))

    def enumerate_hrid(self, marc_record):
        return self.handling == HridHandling.default or "001" not in marc_record

    def get_next_hrid(self, namespace: FOLIONamespaces):
        hrid = ""
        if namespace == FOLIONamespaces.instances:
            hrid = (
                f"{self.instance_hrid_prefix}"
                f"{self.generate_numeric_part(self.instance_hrid_counter)}"
            )
            self.instance_hrid_counter += 1
        elif namespace == FOLIONamespaces.holdings:
            hrid = (
                f"{self.holdings_hrid_prefix}"
                f"{self.generate_numeric_part(self.holdings_hrid_counter)}"
            )
            self.holdings_hrid_counter += 1
        else:
            raise TransformationProcessError("", "Unimplemented namespace")
        return hrid

    def generate_numeric_part(self, counter):
        return str(counter).zfill(11) if self.common_retain_leading_zeroes else str(counter)

    @staticmethod
    def handle_035_generation(
        marc_record: Record,
        legacy_ids,
        migration_report: MigrationReport,
        deactivate035_from001: bool,
        remove_001: bool = True,
    ):
        try:
            f_001 = marc_record["001"].value()
            f_003 = marc_record["003"].value().strip() if "003" in marc_record else ""
            migration_report.add(
                "HridHandling", i18n.t("Values in %{field}", field="003") + f': {f_003 or "Empty"}'
            )

            if deactivate035_from001:
                migration_report.add("HridHandling", i18n.t("035 generation from 001 turned off"))
            else:
                str_035 = f"({f_003}){f_001}" if f_003 else f"{f_001}"
                new_035 = Field(
                    tag="035",
                    indicators=[" ", " "],
                    subfields=[Subfield(code="a", value=str_035)],
                )

                # Don't add the 035 if an identical field already exists
                existing_035 = marc_record.get_fields("035")
                if not any(compare_fields(new_035, e) for e in existing_035):
                    marc_record.add_ordered_field(new_035)
                migration_report.add("HridHandling", i18n.t("Added 035 from 001"))
            if remove_001:
                marc_record.remove_fields("001", "003")

        except Exception:
            if "001" in marc_record:
                s = i18n.n("Failed to create %{to} from %{fro}", to="001", fro="035")
                migration_report.add("HridHandling", s)
                Helper.log_data_issue(legacy_ids, s, marc_record["001"])
            else:
                migration_report.add("HridHandling", i18n.t("Legacy bib records without 001"))

    def hrids_not_updated(self):
        return (
            self.hrid_settings["instances"]["startNumber"] == self.instance_hrid_counter
            and self.hrid_settings["holdings"]["startNumber"] == self.holdings_hrid_counter
            and self.hrid_settings["items"]["startNumber"] == self.items_hrid_counter
        )

    def store_hrid_settings(self):
        logging.info("Setting HRID counter to current")
        try:
            if self.hrids_not_updated():
                logging.info("NOT POSTing HRID settings, since did not change.")
                return

            self.hrid_settings["instances"]["startNumber"] = self.instance_hrid_counter
            self.hrid_settings["holdings"]["startNumber"] = self.holdings_hrid_counter
            self.hrid_settings["items"]["startNumber"] = self.items_hrid_counter
            url = self.folio_client.okapi_url + self.hrid_path
            resp = httpx.put(
                url,
                json=self.hrid_settings,
                headers=self.folio_client.okapi_headers,
            )
            resp.raise_for_status()
            logging.info("%s Successfully set HRID settings.", resp.status_code)
            a = self.folio_client.folio_get_single_object(self.hrid_path)
            logging.info("Current hrid settings: %s", json.dumps(a, indent=4))
        except Exception:
            logging.exception(
                f"Something went wrong when setting the HRID settings. "
                f"Update them manually. {json.dumps(self.hrid_settings)}"
            )

    def reset_instance_hrid_counter(self):
        logging.info("Resetting Instances HRID settings to 1")
        self.instance_hrid_counter = 1
        self.migration_report.set(
            "GeneralStatistics",
            i18n.t("Instances HRID starting number"),
            self.instance_hrid_counter,
        )
        self.store_hrid_settings()

    def reset_holdings_hrid_counter(self):
        logging.info("Resetting Holdings HRID settings to 1")
        self.holdings_hrid_counter = 1
        self.migration_report.set(
            "GeneralStatistics", "Holdings HRID starting number", self.holdings_hrid_counter
        )
        self.store_hrid_settings()

    def reset_item_hrid_counter(self):
        logging.info("Resetting Items HRID settings to 1")
        self.items_hrid_counter = 1
        self.migration_report.set(
            "GeneralStatistics", "Items HRID starting number", self.items_hrid_counter
        )
        self.store_hrid_settings()

    def preserve_001_as_hrid(
        self,
        folio_record: dict,
        marc_record: Record,
        legacy_ids: list[str],
        namespace: FOLIONamespaces,
    ):
        value = marc_record["001"].value()
        if value in self.unique_001s:
            self.migration_report.add(
                "HridHandling",
                i18n.t(
                    "Duplicate 001. Creating HRID instead.\n Previous 001 will be stored in a new 035 field"
                ),
            )
            self.handle_035_generation(
                marc_record, legacy_ids, self.migration_report, self.deactivate035_from001
            )
            Helper.log_data_issue(
                legacy_ids,
                "Duplicate 001 for record. HRID created for record",
                value,
            )
            folio_record["hrid"] = self.get_next_hrid(namespace)
            new_001 = Field(tag="001", data=folio_record["hrid"])
            marc_record.add_ordered_field(new_001)
            self.instance_hrid_counter += 1
        else:
            self.unique_001s.add(value)
            folio_record["hrid"] = value
            self.migration_report.add("HridHandling", i18n.t("Took HRID from 001"))


def compare_fields(field1: Field, field2: Field) -> bool:
    bool_compare = (
        field1.tag == field2.tag
        and field1.indicators == field2.indicators
        and field1.subfields == field2.subfields
    )
    return bool_compare



import logging
import i18n
from datetime import datetime
from datetime import timezone


class MigrationReport:
    """Class responsible for handling the migration report"""

    def __init__(self):
        self.report = {}
        self.stats = {}

    def add(self, blurb_id, measure_to_add, number=1):
        """Add section header and values to migration report.

        Args:
            blurb_id (string): ID of Blurb in translations file
            measure_to_add (_type_): _description_
            number (int, optional): _description_. Defaults to 1.
        """
        try:
            self.report[blurb_id][measure_to_add] += number
        except KeyError:
            if blurb_id not in self.report:
                self.report[blurb_id] = {"blurb_id": blurb_id}
            if measure_to_add not in self.report[blurb_id]:
                self.report[blurb_id][measure_to_add] = number

    def set(self, blurb_id, measure_to_add: str, number: int):
        """Set a section value  to a specific number

        Args:
            blurb (_type_): _description_
            measure_to_add (str): _description_
            number (int): _description_
        """
        if blurb_id not in self.report:
            self.report[blurb_id] = {}
        self.report[blurb_id][measure_to_add] = number

    def add_general_statistics(self, measure_to_add: str):
        """Shortcut for adding to the first breakdown

        Args:
            measure_to_add (str): _description_
        """
        self.add("GeneralStatistics", measure_to_add)

    def write_migration_report(
        self,
        report_title,
        report_file,
        time_started: datetime,
    ):
        """Writes the migration report, including section headers, section blurbs, and values.

        Args:
            report_title (_type_):the header of the report.
            report_file (_type_):path to file
            time_started (datetime): The datetime stamp (in utc), of when the process started
        """
        time_finished = datetime.now(timezone.utc)
        report_file.write(
            "\n".join(
                [
                    "# " + report_title,
                    i18n.t("blurbs.Introduction.description"),
                    "## " + i18n.t("Timings"),
                    "",
                    i18n.t("Measure") + " | " + i18n.t("Value"),
                    "--- | ---:",
                    i18n.t("Time Started:") + " | " + datetime.isoformat(time_started),
                    i18n.t("Time Finished:") + " | " + datetime.isoformat(time_finished),
                    i18n.t("Elapsed time:") + " | " + str(time_finished - time_started),
                ]
            )
        )
        logging.info(f"Elapsed time: {time_finished-time_started}")
        for a in self.report:
            blurb_id = self.report[a].get("blurb_id") or ""
            report_file.write(
                "\n".join(
                    [
                        "",
                        "## " + i18n.t(f"blurbs.{blurb_id}.title"),
                        i18n.t(f"blurbs.{blurb_id}.description"),
                        "<details><summary>"
                        + i18n.t("Click to expand all %{count} things", count=len(self.report[a]))
                        + "</summary>",
                        "",
                        i18n.t("Measure") + " | " + i18n.t("Count"),
                        "--- | ---:",
                    ]
                    + [
                        f"{k or 'EMPTY'} | {self.report[a][k]:,}"
                        for k in sorted(self.report[a], key=as_str)
                        if k != "blurb_id"
                    ]
                    + ["</details>", ""]
                )
            )

    def log_me(self):
        for a in self.report:
            blurb_id = self.report[a].get("blurb_id") or ""
            logging.info(f"{blurb_id}    ")
            logging.info("_______________")
            b = self.report[a]
            sortedlist = [(k, b[k]) for k in sorted(b, key=as_str) if k != "blurb_id"]
            for b in sortedlist:
                logging.info(f"{b[0] or 'EMPTY'} \t\t{b[1]:,}   ")


def as_str(s):
    try:
        return str(s), ""
    except ValueError:
        return "", s



import json
import logging
import sys
from typing import Optional

import i18n
from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.user_mapper import UserMapper
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class UserTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        group_map_path: str
        departments_map_path: Optional[str] = ""
        use_group_map: Optional[bool] = True
        user_mapping_file_name: str
        user_file: FileDefinition
        remove_id_and_request_preferences: Optional[bool] = False

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.users

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        super().__init__(library_config, task_config, use_logging)
        self.task_config = task_config
        self.total_records = 0

        self.user_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder / self.task_config.user_mapping_file_name
        )
        self.folio_keys = []
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(self.user_map)
        # Properties
        self.failed_ids = []
        self.failed_objects = []
        if (
            self.folder_structure.mapping_files_folder / self.task_config.group_map_path
        ).is_file():
            group_mapping = self.load_ref_data_mapping_file(
                "patronGroup",
                self.folder_structure.mapping_files_folder / self.task_config.group_map_path,
                self.folio_keys,
            )
        else:
            logging.info(
                "%s not found. No patronGroup mapping will be performed",
                self.folder_structure.mapping_files_folder / self.task_config.group_map_path,
            )
            group_mapping = []

        if (
            self.folder_structure.mapping_files_folder / self.task_config.departments_map_path
        ).is_file():
            departments_mapping = self.load_ref_data_mapping_file(
                "departments",
                self.folder_structure.mapping_files_folder / self.task_config.departments_map_path,
                self.folio_keys,
            )
        else:
            logging.info(
                "%s not found. No departments mapping will be performed",
                self.folder_structure.mapping_files_folder / self.task_config.departments_map_path,
            )
            departments_mapping = []
        map_path = (
            self.folder_structure.mapping_files_folder / self.task_config.user_mapping_file_name
        )
        with open(map_path, encoding="utf8") as mapping_file:
            user_map = json.load(mapping_file)
            self.mapper = UserMapper(
                self.folio_client,
                task_config,
                library_config,
                user_map,
                departments_mapping,
                group_mapping,
            )

        logging.info("UserTransformer init done")

    def do_work(self):
        logging.info("Starting....")
        source_path = (
            self.folder_structure.legacy_records_folder / self.task_config.user_file.file_name
        )

        try:
            with open(
                self.folder_structure.created_objects_path,
                "w+",
                encoding="utf-8",
            ) as results_file:
                with open(source_path, encoding="utf8") as object_file:
                    logging.info(f"processing {source_path}")
                    file_format = "tsv" if str(source_path).endswith(".tsv") else "csv"
                    for num_users, legacy_user in enumerate(
                        self.mapper.get_users(object_file, file_format), start=1
                    ):
                        try:
                            if num_users == 1:
                                logging.info("First Legacy  user")
                                logging.info(json.dumps(legacy_user, indent=4))
                                print_email_warning()
                            folio_user, index_or_id = self.mapper.do_map(
                                legacy_user,
                                num_users,
                                FOLIONamespaces.users,
                            )
                            folio_user = self.mapper.perform_additional_mapping(
                                legacy_user, folio_user, index_or_id
                            )
                            self.clean_user(folio_user, index_or_id)
                            results_file.write(f"{json.dumps(folio_user)}\n")
                            if num_users == 1:
                                logging.info("## First FOLIO  user")
                                logging.info(json.dumps(folio_user, indent=4, sort_keys=True))
                            self.mapper.migration_report.add_general_statistics(
                                i18n.t("Successful user transformations")
                            )
                            if num_users % 1000 == 0:
                                logging.info(f"{num_users} users processed.")
                        except TransformationRecordFailedError as tre:
                            self.mapper.migration_report.add_general_statistics(
                                i18n.t("Records failed")
                            )
                            Helper.log_data_issue(tre.index_or_id, tre.message, tre.data_value)
                            logging.error(tre)
                        except TransformationProcessError as tpe:
                            logging.critical(tpe)
                            print(f"\n{tpe.message}: {tpe.data_value}")
                            print("\nHalting")
                            sys.exit(1)
                        except ValueError as ve:
                            logging.error(ve)
                            raise ve
                        except Exception as ee:
                            logging.error(ee)
                            logging.error(num_users)
                            logging.error(json.dumps(legacy_user))
                            self.mapper.migration_report.add_general_statistics(
                                i18n.t("Failed user transformations")
                            )
                            logging.error(ee, exc_info=True)

                        self.total_records = num_users
        except FileNotFoundError as fnfe:
            logging.exception("File not found")
            print(f"\n{fnfe}")
            sys.exit(1)

    def wrap_up(self):
        self.extradata_writer.flush()
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Users transformation report"),
                migration_report_file,
                self.mapper.start_datetime,
            )
            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        logging.info("All done!")
        self.clean_out_empty_logs()

    @staticmethod
    def clean_user(folio_user, index_or_id):
        # Make sure the user has exactly one primary address
        if addresses := folio_user.get("personal", {}).get("addresses", []):
            primary_true = []
            for address in addresses:
                if "primaryAddress" not in address:
                    address["primaryAddress"] = False
                elif (
                    isinstance(address["primaryAddress"], bool)
                    and address["primaryAddress"] is True
                ):
                    primary_true.append(address)

            if len(primary_true) < 1:
                addresses[0]["primaryAddress"] = True
            elif len(primary_true) > 1:
                logging.log(
                    26,
                    "DATA ISSUE\t%s\t%s\t%s",
                    index_or_id,
                    "Too many addresses mapped as primary. Setting first one to primary.",
                    primary_true,
                )
                for pt in primary_true[1:]:
                    pt["primaryAddress"] = False


def print_email_warning():
    s = (
        "  ______   __  __              _____   _         _____     ___  \n"  # noqa: E501, W605
        " |  ____| |  \/  |     /\     |_   _| | |       / ____|   |__ \ \n"  # noqa: E501, W605
        " | |__    | \  / |    /  \      | |   | |      | (___        ) |\n"  # noqa: E501, W605
        " |  __|   | |\/| |   / /\ \     | |   | |       \___ \      / / \n"  # noqa: E501, W605
        " |______| |_|  |_| /_/    \_\ |_____| |______| |_____/     (_)  \n"  # noqa: E501, W605
        "                                                                \n"  # noqa: E501, W605
        "                                                       \n"
    )
    print(s)



import csv


class InsensitiveDictReader(csv.DictReader):
    # This class overrides the csv.fieldnames property, which converts all
    # fieldnames without leading and trailing
    # spaces and to lower case.
    @property
    def fieldnames(self):
        return [field.strip().lower() for field in csv.DictReader.fieldnames.fget(self)]

    def next(self):
        return InsensitiveDict(csv.DictReader.next(self))


class InsensitiveDict(dict):
    # This class overrides the __getitem__ method to automatically strip()
    # and lower() the input key
    def __getitem__(self, key):
        return dict.__getitem__(self, key.strip().lower())



import csv
import json
import logging
import sys
import i18n

from dateutil.parser import parse
from folio_uuid.folio_namespaces import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.notes_mapper import NotesMapper
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class UserMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        task_config,
        library_config,
        user_map,
        departments_mapping,
        groups_map,
    ):
        try:
            user_schema = folio_client.get_from_github(
                "folio-org", "mod-user-import", "/ramls/schemas/userdataimport.json"
            )

            user_schema["properties"]["requestPreference"] = folio_client.get_from_github(
                "folio-org", "mod-user-import", "/ramls/schemas/userImportRequestPreference.json"
            )

            super().__init__(
                folio_client,
                user_schema,
                user_map,
                None,
                FOLIONamespaces.users,
                library_config,
            )
            self.task_config = task_config
            self.notes_mapper: NotesMapper = NotesMapper(
                self.library_configuration,
                self.folio_client,
                self.record_map,
                FOLIONamespaces.users,
                True,
            )
            self.notes_mapper.migration_report = self.migration_report
            self.setup_departments_mapping(departments_mapping)
            self.setup_groups_mapping(groups_map)

            for m in self.record_map["data"]:
                if m["folio_field"].startswith("customFields"):
                    if "properties" not in self.schema["properties"]["customFields"]:
                        self.schema["properties"]["customFields"]["properties"] = {}
                    custom_field_prop_name = m["folio_field"].split(".")[-1]
                    self.schema["properties"]["customFields"]["properties"][
                        custom_field_prop_name
                    ] = {"type": "string", "description": "dynamically added custom prop"}

            logging.info("Init done.")
        except TransformationProcessError as tpe:
            logging.critical(tpe)
            print(f"\n{tpe.message}\t{tpe.data_value}")
            sys.exit(1)

    def perform_additional_mapping(self, legacy_user, folio_user, index_or_id):
        self.notes_mapper.map_notes(
            legacy_user, index_or_id, folio_user["id"], FOLIONamespaces.users
        )
        if "personal" not in folio_user:
            folio_user["personal"] = {}
        folio_user["personal"]["preferredContactTypeId"] = "Email"
        folio_user["active"] = True
        if folio_user.get("requestPreference"):
            folio_user["requestPreference"].update(
                {
                    "holdShelf": True,
                    "delivery": False,
                }
            )
        else:
            folio_user["requestPreference"] = {
                "holdShelf": True,
                "delivery": False,
            }

        clean_folio_object = self.validate_required_properties(
            index_or_id, folio_user, self.schema, FOLIONamespaces.users
        )
        if not clean_folio_object.get("personal", {}).get("lastName", ""):
            raise TransformationRecordFailedError(index_or_id, "Last name is missing", "")

        if "preferredFirstName" in clean_folio_object.get(
            "personal", {}
        ) and not clean_folio_object.get("personal", {}).get("preferredFirstName", ""):
            del clean_folio_object["personal"]["preferredFirstName"]

        if self.task_config.remove_id_and_request_preferences:
            del clean_folio_object["id"]
            del clean_folio_object["requestPreference"]
        self.report_folio_mapping_no_schema(clean_folio_object)
        self.report_legacy_mapping_no_schema(legacy_user)

        return clean_folio_object

    def get_users(self, source_file, file_format: str):
        csv.register_dialect("tsv", delimiter="\t")
        if file_format == "tsv":
            reader = csv.DictReader(source_file, dialect="tsv")
        else:  # Assume csv
            reader = csv.DictReader(source_file)
        for idx, row in enumerate(reader):
            if len(row.keys()) < 3:
                raise TransformationProcessError(
                    idx, "something is wrong source file row", json.dumps(row)
                )
            yield row

    def get_prop(self, legacy_user, folio_prop_name, index_or_id, schema_default_value):
        mapped_value = super().get_prop(
            legacy_user, folio_prop_name, index_or_id, schema_default_value
        )
        if folio_prop_name == "personal.addresses.id":
            return ""
        elif folio_prop_name == "patronGroup":
            if self.groups_mapping:
                return self.get_mapped_name(
                    self.groups_mapping,
                    legacy_user,
                    index_or_id,
                    False,
                )
            else:
                return mapped_value
        elif folio_prop_name.startswith("departments"):
            if not self.departments_mapping:
                raise TransformationProcessError(
                    "",
                    "No Departments mapping set up. Set up a departments mapping file "
                    " or remove the mapping of the Departments field",
                )
            return self.get_mapped_name(
                self.departments_mapping,
                legacy_user,
                index_or_id,
                False,
            )
        elif folio_prop_name in ["expirationDate", "enrollmentDate", "personal.dateOfBirth"]:
            return self.get_parsed_date(mapped_value, folio_prop_name)
        return mapped_value

    def get_parsed_date(self, mapped_value, folio_prop_name: str):
        try:
            if not mapped_value.strip():
                return ""
            format_date = parse(mapped_value, fuzzy=True)
            return format_date.isoformat()
        except Exception as ee:
            v = mapped_value
            logging.error(f"{folio_prop_name} {v} could not be parsed: {ee}")
            fmt_string = i18n.t(
                "Parsing error! %{prop_name}: %{value}. The empty string was returned",
                prop_name=folio_prop_name,
                value=v,
            )
            self.migration_report.add("DateTimeConversions", fmt_string)
            return ""

    def setup_groups_mapping(self, groups_map):
        if groups_map:
            self.groups_mapping = RefDataMapping(
                self.folio_client,
                "/groups",
                "usergroups",
                groups_map,
                "group",
                "UserGroupMapping",
            )
        else:
            self.groups_mapping = None

    def setup_departments_mapping(self, departments_mapping):
        if departments_mapping:
            self.departments_mapping = RefDataMapping(
                self.folio_client,
                "/departments",
                "departments",
                departments_mapping,
                "name",
                "DepartmentsMapping",
            )
        else:
            self.departments_mapping = None



import csv
import ctypes
import json
import logging
import sys
import time
import uuid
import i18n
from hashlib import sha1
from os.path import isfile
from typing import List
from typing import Optional

from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.organization_mapper import (
    OrganizationMapper,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration

csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))


# Read files and do some work
class OrganizationTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        files: List[FileDefinition]
        organization_map_path: str
        organization_types_map_path: Optional[str] = ""
        address_categories_map_path: Optional[str] = ""
        email_categories_map_path: Optional[str] = ""
        phone_categories_map_path: Optional[str] = ""

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.organizations

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        csv.register_dialect("tsv", delimiter="\t")

        super().__init__(library_config, task_configuration, use_logging)
        self.object_type_name = self.get_object_type().name
        self.task_configuration = task_configuration
        self.files = self.list_source_files()
        self.total_records = 0

        self.organization_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder
            / self.task_configuration.organization_map_path
        )

        self.results_path = self.folder_structure.created_objects_path
        self.failed_files: List[str] = []
        self.organizations_id_map = self.load_id_map(
            self.folder_structure.organizations_id_map_path
        )

        self.folio_keys = []
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(
            self.organization_map
        )

        self.mapper = OrganizationMapper(
            self.folio_client,
            self.library_configuration,
            self.organization_map,
            self.load_ref_data_mapping_file(
                "organizationTypes",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.organization_types_map_path,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(
                "addresses[0].categories[0]",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.address_categories_map_path,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(
                "emails[0].categories[0]",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.email_categories_map_path,
                self.folio_keys,
                False,
            ),
            self.load_ref_data_mapping_file(
                "phoneNumbers[0].categories[0]",
                self.folder_structure.mapping_files_folder
                / self.task_configuration.phone_categories_map_path,
                self.folio_keys,
                False,
            ),
        )

        self.embedded_extradata_object_cache: set = set()
        self.interfaces_cache: dict = {}

    def list_source_files(self):
        files = [
            self.folder_structure.data_folder / self.object_type_name / f.file_name
            for f in self.task_configuration.files
            if isfile(self.folder_structure.data_folder / self.object_type_name / f.file_name)
        ]
        if not any(files):
            ret_str = ",".join(f.file_name for f in self.task_configuration.files)
            raise TransformationProcessError(
                f"Files {ret_str} not found in"
                "{self.folder_structure.data_folder} / {self.object_type_name}"
            )
        logging.info("Files to process:")
        for filename in files:
            logging.info("\t%s", filename)
        return files

    def process_single_file(self, filename):
        with open(filename, encoding="utf-8-sig") as records_file, open(
            self.folder_structure.created_objects_path, "w+"
        ) as results_file:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of files processed")
            )
            start = time.time()
            records_processed = 0
            for idx, record in enumerate(self.mapper.get_objects(records_file, filename)):
                records_processed += 1
                try:
                    if idx == 0:
                        logging.info("First legacy record:")
                        logging.info(json.dumps(record, indent=4))

                    folio_rec, legacy_id = self.mapper.do_map(
                        record, f"row {idx}", FOLIONamespaces.organizations
                    )
                    self.mapper.report_folio_mapping(folio_rec, self.mapper.organization_schema)

                    # Create extradata and clean the record up
                    folio_rec = self.handle_embedded_extradata_objects(folio_rec)
                    self.mapper.notes_mapper.map_notes(
                        record,
                        legacy_id,
                        folio_rec["id"],
                        FOLIONamespaces.organizations,
                    )
                    folio_rec = self.clean_org(folio_rec)
                    self.organizations_id_map[legacy_id] = self.mapper.get_id_map_tuple(
                        legacy_id, folio_rec, self.object_type
                    )

                    Helper.write_to_file(results_file, folio_rec)

                    if idx == 0:
                        logging.info("First FOLIO record:")
                        logging.info(json.dumps(folio_rec, indent=4))

                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as error:
                    self.mapper.handle_transformation_record_failed_error(idx, error)
                except Exception as excepion:
                    self.mapper.handle_generic_exception(idx, excepion)

                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Number of objects in source data file")
                )
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Number of organizations created")
                )

                # TODO Rewrite to base % value on number of rows in file
                if idx > 1 and idx % 50 == 0:
                    elapsed = idx / (time.time() - start)
                    elapsed_formatted = "{0:.4g}".format(elapsed)
                    logging.info(  # pylint: disable=logging-fstring-interpolation
                        f"{idx:,} records processed. Recs/sec: {elapsed_formatted} "
                    )

            self.total_records = records_processed

            logging.info(  # pylint: disable=logging-fstring-interpolation
                f"Done processing {filename} containing {self.total_records:,} records. "
                f"Total records processed: {self.total_records:,}"
            )

    def do_work(self):
        logging.info("Getting started!")
        for file in self.files:
            logging.info("Processing %s", file)
            try:
                self.process_single_file(file)
            except Exception as ee:
                error_str = (
                    f"Processing of {file} failed:\n{ee}."
                    "Check source files for empty rows or missing reference data"
                )
                logging.exception(error_str)
                self.mapper.migration_report.add("FailedFiles", f"{file} - {ee}")
                sys.exit()

    def wrap_up(self):
        logging.info("Done. Transformer wrapping up...")
        self.extradata_writer.flush()
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            logging.info(
                "Writing migration- and mapping report to %s",
                self.folder_structure.migration_reports_file,
            )
            self.mapper.migration_report.write_migration_report(
                i18n.t("Ogranization transformation report"),
                migration_report_file,
                self.start_datetime,
            )

            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )

            self.mapper.save_id_map_file(
                self.folder_structure.organizations_id_map_path, self.organizations_id_map
            )
        self.clean_out_empty_logs()

        logging.info("All done!")

    def clean_org(self, record):
        if record.get("addresses"):
            self.clean_addresses(record)
        if record.get("interfaces"):
            self.validate_uri(record)

        return record

    def clean_addresses(self, record):
        addresses = record.get("addresses", [])
        primary_address_exists = False
        empty_addresses = []

        for address in addresses:
            # Check if the address has content
            address_content = {k: v for k, v in address.items() if k != "isPrimary"}
            if not any(address_content.values()):
                empty_addresses.append(address)

            # Check if the address is primary
            if address.get("isPrimary") is True:
                primary_address_exists = True

        # If none of the existing addresses is pimrary
        # Make the first one primary
        if not primary_address_exists:
            addresses[0]["isPrimary"] = True

        record["addresses"] = [a for a in addresses if a not in empty_addresses]

        return record

    def validate_uri(self, record):
        valid_interfaces = []
        uri_prefixes = ("ftp://", "sftp://", "http://", "https://")

        for interface in record.get("interfaces"):
            if ("uri" not in interface) or (interface.get("uri", "").startswith(uri_prefixes)):
                valid_interfaces.append(interface)
            else:
                self.mapper.migration_report.add(
                    "MalformedInterfaceUri",
                    i18n.t("Interfaces"),
                )
                Helper.log_data_issue(
                    f"{record['code']}",
                    f"INTERFACE FAILED Malformed interface URI: {interface['uri']}",
                    interface,
                )

        record["interfaces"] = valid_interfaces

        return record

    def handle_embedded_extradata_objects(self, record):
        if record.get("interfaces"):
            extradata_object_type = "interfaces"
            ids_of_external_objects = []

            for embedded_interface in record[extradata_object_type]:
                interface_credential = embedded_interface.pop("interfaceCredential", None)

                interface_id = self.create_referenced_extradata_object(
                    embedded_interface, extradata_object_type
                )
                ids_of_external_objects.append(interface_id)

                if interface_credential and "username" in interface_credential:
                    interface_credential["interfaceId"] = interface_id
                    self.create_referenced_extradata_object(
                        interface_credential, "interfaceCredential"
                    )

            record[extradata_object_type] = ids_of_external_objects

        if record.get("contacts"):
            extradata_object_type = "contacts"
            ids_of_external_objects = []

            for embedded_contact in record[extradata_object_type]:
                if embedded_contact.get("firstName") and embedded_contact.get("lastName"):
                    ids_of_external_objects.append(
                        self.create_referenced_extradata_object(
                            embedded_contact, extradata_object_type
                        )
                    )

            record[extradata_object_type] = ids_of_external_objects

        if "notes" in record:
            # TODO Do the same as for Contacts/Interfaces? Check implementation for Users.
            pass

        return record

    def create_referenced_extradata_object(self, embedded_object, extradata_object_type):
        """Creates an extradata object from an embedded object,
        and returns the UUID.

        Args:
            embedded_object (_type_): _description_
            extradata_object_type (_type_): _description_

        Returns:
            _type_: The organization record with linked extradata UUIDs.
        """
        embedded_object_hash = sha1(
            json.dumps(embedded_object, sort_keys=True).encode("utf-8"), usedforsecurity=False
        ).hexdigest()

        identical_objects = [
            value
            for value in self.embedded_extradata_object_cache
            if value == embedded_object_hash
        ]

        if len(identical_objects) > 0:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of reoccuring identical %{type}", type=extradata_object_type)
            )
            Helper.log_data_issue(
                f"{self.legacy_id}",
                f"Identical {extradata_object_type} objects found in multiple organizations",
                embedded_object,
            )

        extradata_object_uuid = str(uuid.uuid4())
        embedded_object["id"] = extradata_object_uuid

        self.extradata_writer.write(extradata_object_type, embedded_object)
        self.embedded_extradata_object_cache.add(embedded_object_hash)

        self.mapper.migration_report.add_general_statistics(
            i18n.t("Number of linked %{type} created", type=extradata_object_type)
        )

        return extradata_object_uuid



import json
import logging
import os
import re
import sys

import httpx
from folio_uuid.folio_uuid import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.notes_mapper import NotesMapper
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class OrganizationMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        library_configuration: LibraryConfiguration,
        organization_map: dict,
        organization_types_map,
        address_categories_map,
        email_categories_map,
        phone_categories_map,
    ):
        # Build composite organization schema
        if os.environ.get("GITHUB_TOKEN"):
            logging.info("Using GITHB_TOKEN environment variable for Gihub API Access")
        organization_schema = OrganizationMapper.get_latest_acq_schemas_from_github(
            "folio-org", "mod-organizations-storage", "mod-orgs", "organization"
        )

        super().__init__(
            folio_client,
            organization_schema,
            organization_map,
            None,
            FOLIONamespaces.organizations,
            library_configuration,
        )
        self.organization_schema = organization_schema
        # Set up reference data maps
        self.set_up_reference_data_mapping(
            organization_types_map,
            address_categories_map,
            email_categories_map,
            phone_categories_map,
        )

        self.folio_client: FolioClient = folio_client
        self.notes_mapper: NotesMapper = NotesMapper(
            library_configuration,
            self.folio_client,
            organization_map,
            FOLIONamespaces.note,
            True,
        )
        self.notes_mapper.migration_report = self.migration_report

    # Commence the mapping work
    def get_prop(self, legacy_organization, folio_prop_name, index_or_id, schema_default_value):
        value_tuple = (
            legacy_organization,
            index_or_id,
            folio_prop_name,
        )

        # Perfrom reference data mappings
        if folio_prop_name == "organizationTypes":
            return self.get_mapped_ref_data_value(
                self.organization_types_map,
                *value_tuple,
                False,
            )

        elif re.compile("addresses\[(\d+)\]\.categories\[(\d+)\]").fullmatch(folio_prop_name):
            return self.get_mapped_ref_data_value(
                self.address_categories_map,
                *value_tuple,
                False,
            )

        elif re.compile("emails\[(\d+)\]\.categories\[(\d+)\]").fullmatch(folio_prop_name):
            return self.get_mapped_ref_data_value(
                self.email_categories_map,
                *value_tuple,
                False,
            )

        elif re.compile("phoneNumbers\[(\d+)\]\.categories\[(\d+)\]").fullmatch(folio_prop_name):
            return self.get_mapped_ref_data_value(
                self.phone_categories_map,
                *value_tuple,
                False,
            )

        elif re.compile("interfaces\[(\d+)\]\.interfaceCredential.interfaceId").fullmatch(
            folio_prop_name
        ):
            return "replace_with_interface_id"

        return super().get_prop(
            legacy_organization, folio_prop_name, index_or_id, schema_default_value
        )

    def set_up_reference_data_mapping(
        self,
        organization_types_map,
        address_categories_map,
        email_categories_map,
        phone_categories_map,
    ):
        """

        Args:
            organization_types_map (_type_): _description_
            address_categories_map (_type_): _description_
            email_categories_map (_type_): _description_
            phone_categories_map (_type_): _description_
        """

        categories_shared_args = (
            self.folio_client,
            "/organizations-storage/categories",
            "categories",
        )

        if address_categories_map:
            self.address_categories_map = RefDataMapping(
                *categories_shared_args, address_categories_map, "value", "CategoriesMapping"
            )
        else:
            self.address_categories_map = None

        if email_categories_map:
            self.email_categories_map = RefDataMapping(
                *categories_shared_args, email_categories_map, "value", "CategoriesMapping"
            )
        else:
            self.email_categories_map = None

        if phone_categories_map:
            self.phone_categories_map = RefDataMapping(
                *categories_shared_args, phone_categories_map, "value", "CategoriesMapping"
            )
        else:
            self.phone_categories_map = None

        if organization_types_map:
            self.organization_types_map = RefDataMapping(
                self.folio_client,
                "/organizations-storage/organization-types",
                "organizationTypes",
                organization_types_map,
                "name",
                "OrganizationTypeMapping",
            )
        else:
            self.organization_types_map = None

    @staticmethod
    def get_latest_acq_schemas_from_github(owner, repo, module, object):
        """
        Given a repository owner, a repository, a module name and the name
        of a FOLIO acquisition object, returns a schema for that object that
        also includes the schemas of any other referenced acq objects.

        Args:
            owner (_type_): _description_
            repo (_type_): _description_
            module (_type_): _description_
            object (_type_): _description_

        Returns:
            _type_: _description_
        """
        try:
            # Authenticate when calling GitHub, using an API key stored in .env
            github_headers = {
                "content-type": "application/json",
                "User-Agent": "FOLIO Migration Tools (https://github.com/FOLIO-FSE/folio_migration_tools/)",  # noqa:E501,B950
            }

            if os.environ.get("GITHUB_TOKEN"):
                github_headers["authorization"] = f"token {os.environ.get('GITHUB_TOKEN')}"

            # Start talkign to GitHub...
            github_path = "https://raw.githubusercontent.com"
            submodules = OrganizationMapper.get_submodules_of_latest_release(
                owner, repo, github_headers
            )

            # Get the sha's of sunmodules acq-models and raml_utils
            acq_models_sha = next(
                (item["sha"] for item in submodules if item["path"] == "acq-models")
            )

            # # TODO Maybe - fetch raml_utils schemas if deemed necessary
            # raml_utils_sha = next((item["sha"] for item in submodules
            # if item["path"] == "raml-utils"))

            acq_models_path = f"{github_path}/{owner}/acq-models/{acq_models_sha}/{module}/schemas"

            req = httpx.get(f"{acq_models_path}/{object}.json", headers=github_headers)
            req.raise_for_status()

            object_schema = json.loads(req.text)

            # Fetch referenced schemas
            extended_object_schema = OrganizationMapper.build_extended_object(
                object_schema, acq_models_path, github_headers
            )

            return extended_object_schema

        except httpx.HTTPError as http_error:
            logging.critical(f"Halting! \t{http_error}")
            sys.exit(2)

        except json.decoder.JSONDecodeError as json_error:
            logging.critical(json_error)
            sys.exit(2)

    @staticmethod
    def get_submodules_of_latest_release(owner, repo, github_headers):
        """
        Given a repository owner and a repository, identifies the latest
        release of the repository and returns the submodules associated with
        this release.

        Args:
            owner (_type_): _description_
            repo (_type_): _description_
            github_headers (_type_): _description_

        Returns:
            _type_: _description_
        """

        github_path = "https://api.github.com/repos"

        # Get metadata for the latest release
        latest_release_path = f"{github_path}/{owner}/{repo}/releases/latest"
        req = httpx.get(f"{latest_release_path}", headers=github_headers, timeout=None)
        req.raise_for_status()
        latest_release = json.loads(req.text)

        # Get the tag assigned to the latest release
        release_tag = latest_release["tag_name"]
        logging.info(f"Using schemas from latest {repo} release: {release_tag}")

        # Get the tree for the latest release
        tree_path = f"{github_path}/{owner}/{repo}/git/trees/{release_tag}"
        req = httpx.get(tree_path, headers=github_headers, timeout=None)
        req.raise_for_status()
        release_tree = json.loads(req.text)

        # Loop through the tree to get the sha of the folder with path "ramls"
        ramls_sha = next((item["sha"] for item in release_tree["tree"] if item["path"] == "ramls"))

        # Get the tree for the ramls folder
        ramls_path = f"{github_path}/{owner}/{repo}/git/trees/{ramls_sha}"
        req = httpx.get(ramls_path, headers=github_headers, timeout=None)
        req.raise_for_status()
        ramls_tree = json.loads(req.text)

        # Loop through the tree to get the sha of submodules
        submodules = [item for item in ramls_tree["tree"] if item["mode"] == "160000"]

        return submodules

    @staticmethod
    def build_extended_object(object_schema, submodule_path, github_headers):
        """
        Takes an object schema (for example an organization) and the path to a
        submodule repository and returns the same schema with the full schemas
        of subordinate objects (for example aliases).

        Args:
            object_schema (_type_): _description_
            submodule_path (_type_): _description_
            github_headers (_type_): _description_

        Returns:
            _type_: _description_
        """

        supported_types = ["string", "boolean", "number", "integer", "text", "object", "array"]

        try:
            for property_name_level1, property_level1 in object_schema["properties"].items():
                # For now, treat references to UUIDs like strings
                # It's not great practice, but it's the way FOLIO mostly handles it
                if "../../common/schemas/uuid.json" in property_level1.get("$ref", ""):
                    property_level1["type"] = "string"

                # Preliminary implementation of tags
                # https://github.com/folio-org/raml/blob/master/schemas/tags.schema
                elif property_name_level1 == "tags":
                    property_level1["properties"] = {
                        "tagList": {
                            "description": "List of tags",
                            "type": "array",
                            "items": {"type": "string"},
                        }
                    }

                elif property_name_level1 == "contacts":
                    contact_schema = OrganizationMapper.fetch_additional_schema("contact")
                    property_level1["items"] = contact_schema

                elif property_name_level1 == "interfaces":
                    interface_schema = OrganizationMapper.fetch_additional_schema("interface")
                    interface_schema["required"] = ["name"]

                    # Temporarily add the credential object as a subproperty
                    interface_credential_schema = OrganizationMapper.fetch_additional_schema(
                        "interface_credential"
                    )
                    interface_credential_schema["required"] = (
                        ["username", "password", "interfaceId"],
                    )

                    interface_schema["properties"][
                        "interfaceCredential"
                    ] = interface_credential_schema

                    property_level1["items"] = interface_schema

                elif (
                    property_level1.get("type") == "array"
                    and property_level1.get("items").get("$ref")
                    == "../../common/schemas/uuid.json"
                ):
                    property_level1["items"]["type"] = "string"

                # Report and discard unhandled properties
                elif (
                    property_level1.get("type") not in supported_types
                    or "../" in property_level1.get("$ref", "")
                    or "../" in property_level1.get("items", {}).get("$ref", "")
                ):
                    logging.info(f"Property not yet supported: {property_name_level1}")

                # Handle object properties
                elif property_level1.get("type") == "object" and property_level1.get("$ref"):
                    logging.info("Fecthing referenced schema for object %s", property_name_level1)

                    ref_object = property_level1["$ref"]
                    schema_url = f"{submodule_path}/{ref_object}"

                    req = httpx.get(schema_url, headers=github_headers, timeout=None)
                    req.raise_for_status()

                    property_level1 = dict(property_level1, **json.loads(req.text))

                elif property_level1.get("type") == "array" and property_level1.get("items").get(
                    "$ref"
                ):
                    ref_object = property_level1["items"]["$ref"]
                    schema_url = f"{submodule_path}/{ref_object}"

                    req = httpx.get(schema_url, headers=github_headers, timeout=None)
                    req.raise_for_status()

                    property_level1["items"] = dict(
                        property_level1["items"], **json.loads(req.text)
                    )

            return object_schema

        except httpx.HTTPError as he:
            logging.error(he)

    @staticmethod
    def fetch_additional_schema(folio_object):
        additional_schema = OrganizationMapper.get_latest_acq_schemas_from_github(
            "folio-org", "mod-organizations-storage", "mod-orgs", folio_object
        )
        return additional_schema



import logging
import i18n

from folio_uuid.folio_namespaces import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)


class NotesMapper(MappingFileMapperBase):
    def __init__(
        self,
        library_configuration: LibraryConfiguration,
        folio_client: FolioClient,
        record_map: dict,
        object_type: FOLIONamespaces,
        ignore_legacy_identifier: bool = False,
    ) -> None:
        self.folio_client: FolioClient = folio_client
        self.setup_notes_schema()
        super().__init__(
            folio_client,
            self.notes_schema,
            record_map,
            None,
            object_type,
            library_configuration,
            ignore_legacy_identifier,
        )

        self.noteprops = {
            "data": [p for p in record_map["data"] if p["folio_field"].startswith("notes[")]
        }
        logging.info("Set %s props used for note mapping", len(self.noteprops["data"]))
        logging.info("Initiated mapper for Notes")

    def setup_notes_schema(self):
        notes_schemas = self.get_notes_schema()
        self.notes_schema = notes_schemas["noteCollection"]
        self.notes_schema["properties"]["notes"]["items"] = notes_schemas["note"]
        self.notes_schema["required"] = []

    def map_notes(self, legacy_object, legacy_id, object_uuid: str, record_type: FOLIONamespaces):
        if any(self.noteprops["data"]):
            for note in self.do_map(legacy_object, legacy_id, FOLIONamespaces.note)[0].get(
                "notes", []
            ):
                if note.get("content", "").strip():
                    type_string = {
                        FOLIONamespaces.users: "user",
                        FOLIONamespaces.course: "course",
                        FOLIONamespaces.organizations: "organization",
                        FOLIONamespaces.orders: "orders",
                    }.get(record_type)
                    note["links"] = [{"id": object_uuid, "type": type_string}]
                    if "type" in note:
                        del note["type"]
                    self.extradata_writer.write("notes", note)
                    self.migration_report.add_general_statistics(
                        i18n.t("Number of linked notes created")
                    )
                    self.migration_report.add("MappedNoteTypes", note["typeId"])
                else:
                    self.migration_report.add_general_statistics(
                        i18n.t("Number of discarded notes with no content")
                    )

    def get_notes_schema(self):
        notes_schema = self.folio_client.get_from_github(
            "folio-org",
            "mod-notes",
            "src/main/resources/swagger.api/schemas/note.yaml",
        )
        notes_common = self.folio_client.get_from_github(
            "folio-org",
            "mod-notes",
            "src/main/resources/swagger.api/schemas/common.yaml",
        )
        for prop in notes_schema["note"]["properties"].items():
            if prop[1].get("$ref", "") == "common.yaml#/uuid":
                prop[1]["type"] = notes_common["uuid"]["type"]

        for p in ["links", "metadata", "id"]:
            del notes_schema["note"]["properties"][p]
        return notes_schema



import logging
from typing import Annotated
from typing import List
import i18n

from folio_uuid.folio_namespaces import FOLIONamespaces
from pydantic import Field

from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.library_configuration import IlsFlavour
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.marc_rules_transformation.marc_file_processor import (
    MarcFileProcessor,
)
from folio_migration_tools.marc_rules_transformation.rules_mapper_bibs import (
    BibsRulesMapper,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class BibsTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: Annotated[
            str,
            Field(
                description=(
                    "Name of this migration task. The name is being used to call the specific "
                    "task, and to distinguish tasks of similar types"
                )
            ),
        ]
        migration_task_type: Annotated[
            str,
            Field(
                title="Migration task type",
                description=("The type of migration task you want to perform."),
            ),
        ]
        files: Annotated[
            List[FileDefinition],
            Field(
                title="Source files",
                description=("List of MARC21 files with bibliographic records."),
            ),
        ]
        ils_flavour: Annotated[
            IlsFlavour,
            Field(
                title="ILS flavour", description="The type of ILS you are migrating records from."
            ),
        ]
        custom_bib_id_field: Annotated[
            str,
            Field(
                title="Custom BIB ID field",
                description=(
                    'A string representing a MARC field with optional subfield indicated by a "$" '
                    '(eg. "991$a") from which to draw legacy Bib ID. Use this in combination '
                    'with `ilsFlavour: "custom"`. Defaults to "001", and is ignored for all other '
                    "ilsFlavours."
                ),
            ),
        ] = "001"
        add_administrative_notes_with_legacy_ids: Annotated[
            bool,
            Field(
                title="Add administrative notes with legacy IDs",
                description=(
                    "If set to true, an Administrative note will be added to the records "
                    "containing the legacy ID. Use this in order to protect the values from "
                    "getting overwritten by overlays,"
                ),
            ),
        ] = True
        tags_to_delete: Annotated[
            List[str],
            Field(
                title="Tags to delete from MARC record",
                description=(
                    "Tags in the incoming MARC authority that the process should remove "
                    "before adding them into FOLIO. These tags will be used in the "
                    "transformation before getting removed."
                ),
            ),
        ] = []
        create_source_records: Annotated[
            bool,
            Field(
                title="Create source records",
                description=(
                    "Controls wheter or not to retain the MARC records in "
                    "Source Record Storage."
                ),
            ),
        ] = True
        parse_cataloged_date: Annotated[
            bool,
            Field(
                title="Parse cataloged date",
                description=(
                    "Parse fields mapped to catalogedDate into a FOLIO accepted date string using "
                    "dateutil.parser. Verify results carefully when using"
                ),
            ),
        ] = False
        hrid_handling: Annotated[
            HridHandling,
            Field(
                title="HRID Handling",
                description=(
                    "Setting to default will make FOLIO generate HRIDs and move the existing "
                    "001:s into a 035, concatenated with the 003. Choosing preserve001 means "
                    "the 001:s will remain in place, and that they will also become the HRIDs"
                ),
            ),
        ] = HridHandling.default
        reset_hrid_settings: Annotated[
            bool,
            Field(
                title="Reset HRID settings",
                description=(
                    "Setting to true means the task will "
                    "reset the HRID counters for this particular record type"
                ),
            ),
        ] = False
        update_hrid_settings: Annotated[
            bool,
            Field(
                title="Update HRID settings",
                description="At the end of the run, update FOLIO with the HRID settings",
            ),
        ] = True
        deactivate035_from001: Annotated[
            bool,
            Field(
                title="Create 035 from 001 and 003",
                description=(
                    "This deactivates the FOLIO default functionality of moving the previous 001 "
                    "into a 035, prefixed with the value from 003"
                ),
            ),
        ] = False

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.instances

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        super().__init__(library_config, task_config, use_logging)
        self.processor: MarcFileProcessor
        self.check_source_files(
            self.folder_structure.legacy_records_folder, self.task_configuration.files
        )
        self.mapper = BibsRulesMapper(self.folio_client, library_config, self.task_configuration)
        self.bib_ids: set = set()
        if (
            self.task_configuration.reset_hrid_settings
            and self.task_configuration.update_hrid_settings
        ):
            self.mapper.hrid_handler.reset_instance_hrid_counter()
        logging.info("Init done")

    def do_work(self):
        self.do_work_marc_transformer()

    def wrap_up(self):
        logging.info("Done. Transformer wrapping up...")
        self.extradata_writer.flush()
        self.processor.wrap_up()
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Bibliographic records transformation report"),
                report_file,
                self.start_datetime,
            )
            Helper.print_mapping_report(
                report_file,
                self.mapper.parsed_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )

        logging.info(
            "Done. Transformation report written to %s",
            self.folder_structure.migration_reports_file.name,
        )
        self.clean_out_empty_logs()



import csv
import json
import logging
import sys
import time
import traceback
import i18n
from typing import Optional

from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.courses_mapper import (
    CoursesMapper,
)
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration


class CoursesMigrator(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        composite_course_map_path: str
        migration_task_type: str
        courses_file: FileDefinition
        terms_map_path: str
        departments_map_path: str
        look_up_instructor: Optional[bool] = False

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.course

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        self.task_configuration = task_configuration
        super().__init__(library_config, task_configuration)
        self.t0 = time.time()
        self.courses_map = self.setup_records_map(
            self.folder_structure.mapping_files_folder
            / self.task_configuration.composite_course_map_path
        )
        self.folio_keys = MappingFileMapperBase.get_mapped_folio_properties_from_map(
            self.courses_map
        )
        terms_map = self.load_ref_data_mapping_file(
            "terms",
            self.folder_structure.mapping_files_folder / self.task_configuration.terms_map_path,
            self.folio_keys,
        )

        departments_map = self.load_ref_data_mapping_file(
            "departments",
            self.folder_structure.mapping_files_folder
            / self.task_configuration.departments_map_path,
            self.folio_keys,
        )
        self.mapper: CoursesMapper = CoursesMapper(
            self.folio_client,
            self.courses_map,
            terms_map,
            departments_map,
            self.library_configuration,
            self.task_configuration,
        )
        logging.info("Init completed")

    def do_work(self):
        logging.info("Starting")
        full_path = (
            self.folder_structure.legacy_records_folder
            / self.task_configuration.courses_file.file_name
        )
        logging.info("Processing %s", full_path)
        start = time.time()
        with open(full_path, encoding="utf-8-sig") as records_file:
            for idx, record in enumerate(self.mapper.get_objects(records_file, full_path)):
                try:
                    if idx == 0:
                        logging.info("First legacy record:")
                        logging.info(json.dumps(record, indent=4))
                        self.mapper.verify_legacy_record(record, idx)
                    folio_rec, legacy_id = self.mapper.do_map(
                        record, f"row {idx}", FOLIONamespaces.course
                    )
                    self.mapper.perform_additional_mappings((folio_rec, legacy_id))
                    if idx == 0:
                        logging.info("First FOLIO record:")
                        logging.info(json.dumps(folio_rec, indent=4))
                    self.mapper.store_objects((folio_rec, legacy_id))
                    self.mapper.notes_mapper.map_notes(
                        record, legacy_id, folio_rec["course"]["id"], FOLIONamespaces.course
                    )

                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as data_error:
                    self.mapper.handle_transformation_record_failed_error(idx, data_error)
                except AttributeError as attribute_error:
                    traceback.print_exc()
                    logging.fatal(attribute_error)
                    logging.info("Quitting...")
                    sys.exit(1)
                except Exception as excepion:
                    self.mapper.handle_generic_exception(idx, excepion)
                self.mapper.migration_report.add(
                    "GeneralStatistics",
                    i18n.t("Number of Legacy items in %{container}", container=full_path),
                )
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Number of Legacy items in total")
                )
                self.print_progress(idx, start)

    def wrap_up(self):
        self.extradata_writer.flush()
        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Courses migration report"), report_file, self.mapper.start_datetime
            )
        self.clean_out_empty_logs()


def timings(t0, t0func, num_objects):
    avg = (time.time() - t0) / num_objects
    elapsed = time.time() - t0
    elapsed_func = time.time() - t0func
    return (
        f"Total objects: {num_objects}\tTotal elapsed: {elapsed:.2f}\t"
        f"Average per object: {avg:.2f}s\tElapsed this time: {elapsed_func:.2f}"
    )



import logging
import sys
import time
from pathlib import Path

from folio_uuid.folio_namespaces import FOLIONamespaces


class FolderStructure:
    def __init__(
        self,
        base_path: Path,
        object_type: FOLIONamespaces,
        migration_task_name: str,
        iteration_identifier: str,
        add_time_stamp_to_file_names: bool,
    ):
        logging.info("Validating folder structure")

        self.object_type: FOLIONamespaces = object_type
        self.migration_task_name = migration_task_name
        self.add_time_stamp_to_file_names = add_time_stamp_to_file_names
        self.iteration_identifier = iteration_identifier
        self.base_folder = Path(base_path)
        if not self.base_folder.is_dir():
            logging.critical("Base Folder Path is not a folder. Exiting.")
            sys.exit(1)

        # Basic folders
        self.mapping_files_folder = self.base_folder / "mapping_files"
        self.verify_folder(self.mapping_files_folder)
        gitignore = self.base_folder / ".gitignore"
        verify_git_ignore(gitignore)
        self.verify_folder(self.base_folder / "iterations")

        # Iteration-specific folders
        self.iteration_folder = self.base_folder / "iterations" / self.iteration_identifier
        self.verify_folder(self.iteration_folder)
        self.data_folder = self.iteration_folder / "source_data"
        self.verify_folder(self.data_folder)
        self.results_folder = self.iteration_folder / "results"
        self.verify_folder(self.results_folder)
        self.reports_folder = self.iteration_folder / "reports"
        self.verify_folder(self.reports_folder)

    def log_folder_structure(self):
        logging.info("Mapping files folder is %s", self.mapping_files_folder)
        logging.info("Git ignore is set up correctly")
        logging.info("Base folder is %s", self.base_folder)
        logging.info("Reports and logs folder is %s", self.reports_folder)
        logging.info("Results folder is %s", self.results_folder)
        logging.info("Data folder is %s", self.data_folder)
        logging.info("Source records files folder is %s", self.legacy_records_folder)
        logging.info("Log file will be located at %s", self.transformation_log_path)
        logging.info("Extra data will be stored at%s", self.transformation_extra_data_path)
        logging.info("Data issue reports %s", self.data_issue_file_path)
        logging.info("Created objects will be stored at  %s", self.created_objects_path)
        logging.info("Migration report file will be saved at %s", self.migration_reports_file)

    def setup_migration_file_structure(self, source_file_type: str = ""):
        self.time_stamp = f'_{time.strftime("%Y%m%d-%H%M%S")}'
        self.time_str = self.time_stamp if self.add_time_stamp_to_file_names else ""
        self.file_template = f"{self.time_str}_{self.migration_task_name}"
        object_type_string = str(self.object_type.name).lower()
        if source_file_type:
            self.legacy_records_folder = self.data_folder / source_file_type
        elif self.object_type == FOLIONamespaces.other:
            self.legacy_records_folder = self.data_folder
        else:
            self.legacy_records_folder = self.data_folder / object_type_string
        self.verify_folder(self.legacy_records_folder)

        # Make sure the items are there if the Holdings processor is run
        if self.object_type == FOLIONamespaces.holdings:
            self.verify_folder(self.data_folder / str(FOLIONamespaces.items.name).lower())

        self.transformation_log_path = self.reports_folder / (
            f"log_{object_type_string}{self.file_template}.log"
        )

        self.failed_recs_path = (
            self.results_folder / f"failed_records{self.file_template}{self.time_stamp}.txt"
        )

        self.transformation_extra_data_path = (
            self.results_folder / f"extradata{self.file_template}.extradata"
        )

        self.data_issue_file_path = (
            self.reports_folder / f"data_issues_log{self.file_template}.tsv"
        )
        self.created_objects_path = (
            self.results_folder / f"folio_{object_type_string}{self.file_template}.json"
        )
        self.failed_marc_recs_file = (
            self.results_folder / f"failed_records{self.file_template}.mrc"
        )

        self.migration_reports_file = self.reports_folder / f"report{self.file_template}.md"

        self.srs_records_path = (
            self.results_folder / f"folio_srs_{object_type_string}{self.file_template}.json"
        )
        self.organizations_id_map_path = (
            self.results_folder / f"{str(FOLIONamespaces.organizations.name).lower()}_id_map.json"
        )
        self.instance_id_map_path = (
            self.results_folder / f"{str(FOLIONamespaces.instances.name).lower()}_id_map.json"
        )
        self.auth_id_map_path = (
            self.results_folder / f"{str(FOLIONamespaces.authorities.name).lower()}_id_map.json"
        )

        self.holdings_id_map_path = (
            self.results_folder / f"{str(FOLIONamespaces.holdings.name).lower()}_id_map.json"
        )
        self.id_map_path = (
            self.results_folder / f"{str(self.object_type.name).lower()}_id_map.json"
        )
        # Mapping files
        self.material_type_map_path = self.mapping_files_folder / "material_types.tsv"
        self.loan_type_map_path = self.mapping_files_folder / "loan_types.tsv"
        self.temp_loan_type_map_path = self.mapping_files_folder / "temp_loan_types.tsv"
        self.statistical_codes_map_path = self.mapping_files_folder / "statcodes.tsv"
        self.item_statuses_map_path = self.mapping_files_folder / "item_statuses.tsv"

    def verify_folder(self, folder_path: Path):
        if not folder_path.is_dir():
            logging.critical("There is no folder located at %s. Exiting.", folder_path)
            logging.critical("Create a folder by calling\n\tmkdir %s", folder_path)
            sys.exit(1)
        else:
            logging.info("Located %s", folder_path)


def verify_git_ignore(gitignore: Path):
    with open(gitignore, "r+") as f:
        contents = f.read()
        if "results/" not in contents:
            f.write("results/\n")
        if "archive/" not in contents:
            f.write("archive/\n")
        if "source_data/" not in contents:
            f.write("source_data/\n")
        if "*.data" not in contents:
            f.write("*.data\n")
    logging.info("Made sure there was a valid .gitignore file at %s", gitignore)



import csv
import ctypes
import json
import logging
import sys
import time
import traceback
import i18n
from typing import Annotated
from typing import List
from typing import Optional

from folio_uuid.folio_namespaces import FOLIONamespaces
from httpx import HTTPError
from pydantic import Field

from folio_migration_tools.custom_exceptions import TransformationProcessError
from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.holdings_helper import HoldingsHelper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import HridHandling
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.holdings_mapper import (
    HoldingsMapper,
)
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.marc_rules_transformation.hrid_handler import HRIDHandler
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration

csv.field_size_limit(int(ctypes.c_ulong(-1).value // 2))
csv.register_dialect("tsv", delimiter="\t")


class HoldingsCsvTransformer(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        hrid_handling: HridHandling
        files: List[FileDefinition]
        holdings_map_file_name: str
        location_map_file_name: str
        default_call_number_type_name: str
        previously_generated_holdings_files: Optional[list[str]] = []
        fallback_holdings_type_id: str
        holdings_type_uuid_for_boundwiths: Annotated[
            str,
            Field(
                title="Holdings Type for Boundwith Holdings",
                description=(
                    "UUID for a Holdings type (set in Settings->Inventory) "
                    "for Bound-with Holdings)"
                ),
            ),
        ] = ""
        call_number_type_map_file_name: Optional[str]
        holdings_merge_criteria: Optional[list[str]] = [
            "instanceId",
            "permanentLocationId",
            "callNumber",
        ]
        reset_hrid_settings: Optional[bool] = False
        update_hrid_settings: Annotated[
            bool,
            Field(
                title="Update HRID settings",
                description="At the end of the run, update FOLIO with the HRID settings",
            ),
        ] = True

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.holdings

    def __init__(
        self,
        task_config: TaskConfiguration,
        library_config: LibraryConfiguration,
        use_logging: bool = True,
    ):
        super().__init__(library_config, task_config, use_logging)
        self.fallback_holdings_type = None
        try:
            self.task_config = task_config
            self.bound_with_keys = set()
            self.mapper = HoldingsMapper(
                self.folio_client,
                self.load_mapped_fields(),
                self.load_location_map(),
                self.load_call_number_type_map(),
                self.load_id_map(self.folder_structure.instance_id_map_path, True),
                library_config,
            )
            self.holdings = {}
            self.total_records = 0
            self.holdings_id_map = self.load_id_map(self.folder_structure.holdings_id_map_path)
            self.holdings_sources = self.get_holdings_sources()
            self.results_path = self.folder_structure.created_objects_path
            self.holdings_types = list(
                self.folio_client.folio_get_all("/holdings-types", "holdingsTypes")
            )
            logging.info("%s\tholdings types in tenant", len(self.holdings_types))
            self.validate_merge_criterias()
            self.check_source_files(
                self.folder_structure.data_folder / "items", self.task_config.files
            )
            self.fallback_holdings_type = next(
                h
                for h in self.holdings_types
                if h["id"] == self.task_config.fallback_holdings_type_id
            )
            if not self.fallback_holdings_type:
                raise TransformationProcessError(
                    "",
                    (
                        "Holdings type with ID "
                        f"{self.task_config.fallback_holdings_type_id} "
                        "not found in FOLIO."
                    ),
                )
            logging.info(
                "%s will be used as default holdings type",
                self.fallback_holdings_type["name"],
            )
            if any(self.task_config.previously_generated_holdings_files):
                for file_name in self.task_config.previously_generated_holdings_files:
                    logging.info("Processing %s", file_name)
                    self.holdings.update(
                        HoldingsHelper.load_previously_generated_holdings(
                            self.folder_structure.results_folder / file_name,
                            self.task_config.holdings_merge_criteria,
                            self.mapper.migration_report,
                            self.task_config.holdings_type_uuid_for_boundwiths,
                        )
                    )

            else:
                logging.info("No file of legacy holdings setup.")

            if (
                self.task_configuration.reset_hrid_settings
                and self.task_configuration.update_hrid_settings
            ):
                hrid_handler = HRIDHandler(
                    self.folio_client, HridHandling.default, self.mapper.migration_report, True
                )
                hrid_handler.reset_holdings_hrid_counter()

        except HTTPError as http_error:
            logging.critical(http_error)
            sys.exit(1)
        except (FileNotFoundError, TransformationProcessError) as process_error:
            logging.critical(process_error)
            logging.critical("Halting.")
            sys.exit(1)
        except json.JSONDecodeError as jde:
            raise jde
        except Exception as exception:
            logging.info("\n=======ERROR===========")
            logging.info(exception)
            logging.info("\n=======Stack Trace===========")
            traceback.print_exc()
            sys.exit(1)
        logging.info("Init done")

    def load_call_number_type_map(self):
        with open(
            self.folder_structure.mapping_files_folder
            / self.task_config.call_number_type_map_file_name,
            "r",
        ) as callnumber_type_map_f:
            return self.load_ref_data_map_from_file(
                callnumber_type_map_f, "Found %s rows in call number type map"
            )

    def load_location_map(self):
        with open(
            self.folder_structure.mapping_files_folder / self.task_config.location_map_file_name
        ) as location_map_f:
            return self.load_ref_data_map_from_file(
                location_map_f, "Found %s rows in location map"
            )

    # TODO Rename this here and in `load_call_number_type_map` and `load_location_map`
    def load_ref_data_map_from_file(self, file, message):
        ref_dat_map = list(csv.DictReader(file, dialect="tsv"))
        logging.info(message, len(ref_dat_map))
        return ref_dat_map

    def load_mapped_fields(self):
        with open(
            self.folder_structure.mapping_files_folder / self.task_config.holdings_map_file_name
        ) as holdings_mapper_f:
            holdings_map = json.load(holdings_mapper_f)
            logging.info("%s fields in holdings mapping file map", len(holdings_map["data"]))
            mapped_fields = MappingFileMapperBase.get_mapped_folio_properties_from_map(
                holdings_map
            )
            logging.info(
                "%s mapped fields in holdings mapping file map",
                len(list(mapped_fields)),
            )
            return holdings_map

    def load_instance_id_map(self):
        res = {}
        with open(self.folder_structure.instance_id_map_path, "r") as instance_id_map_file:
            for index, json_string in enumerate(instance_id_map_file):
                # Format:{"legacy_id", "folio_id","instanceLevelCallNumber"}
                if index % 500000 == 0:
                    print(f"{index} instance ids loaded to map", end="\r")
                map_object = json.loads(json_string)
                res[map_object["legacy_id"]] = map_object
        logging.info("Loaded %s migrated instance IDs", (index + 1))
        return res

    def do_work(self):
        logging.info("Starting....")
        for file_def in self.task_config.files:
            logging.info("Processing %s", file_def.file_name)
            try:
                self.process_single_file(file_def)
            except Exception as ee:
                error_str = (
                    f"Processing of {file_def.file_name} failed:\n{ee}."
                    "\nCheck source files for empty rows or missing reference data"
                )
                logging.critical(error_str)
                print(f"\n{error_str}\nHalting")
                sys.exit(1)
        logging.info(
            f"processed {self.total_records:,} records in {len(self.task_config.files)} files"
        )

    def wrap_up(self):
        logging.info("Done. Transformer wrapping up...")
        self.extradata_writer.flush()
        if any(self.holdings):
            logging.info(
                "Saving holdings created to %s",
                self.folder_structure.created_objects_path,
            )
            with open(self.folder_structure.created_objects_path, "w+") as holdings_file:
                for holding in self.holdings.values():
                    for legacy_id in holding["formerIds"]:
                        # Prevent the first item in a boundwith to be overwritten
                        # TODO: Find out why not
                        # if legacy_id not in self.holdings_id_map:
                        self.holdings_id_map[legacy_id] = self.mapper.get_id_map_tuple(
                            legacy_id, holding, self.object_type
                        )
                    Helper.write_to_file(holdings_file, holding)
                    self.mapper.migration_report.add_general_statistics(
                        i18n.t("Holdings Records Written to disk")
                    )
            self.mapper.save_id_map_file(
                self.folder_structure.holdings_id_map_path, self.holdings_id_map
            )
        with open(self.folder_structure.migration_reports_file, "w") as migration_report_file:
            self.mapper.migration_report.write_migration_report(
                i18n.t("Holdings transformation report"),
                migration_report_file,
                self.mapper.start_datetime,
            )
            Helper.print_mapping_report(
                migration_report_file,
                self.total_records,
                self.mapper.mapped_folio_fields,
                self.mapper.mapped_legacy_fields,
            )
        logging.info("All done!")
        self.clean_out_empty_logs()

    def validate_merge_criterias(self):
        holdings_schema = self.folio_client.get_holdings_schema()
        properties = holdings_schema["properties"].keys()
        logging.info(properties)
        logging.info(self.task_config.holdings_merge_criteria)
        res = [mc for mc in self.task_config.holdings_merge_criteria if mc not in properties]
        if any(res):
            logging.critical(
                (
                    "Merge criteria(s) is not a property of a holdingsrecord: %s"
                    "check the merge criteria names and try again"
                ),
                ", ".join(res),
            )
            sys.exit(1)

    def process_single_file(self, file_def: FileDefinition):
        full_path = self.folder_structure.data_folder / "items" / file_def.file_name
        with open(full_path, encoding="utf-8-sig") as records_file:
            self.mapper.migration_report.add_general_statistics(
                i18n.t("Number of files processed")
            )
            start = time.time()
            records_processed = 0
            for idx, legacy_record in enumerate(self.mapper.get_objects(records_file, full_path)):
                records_processed = idx + 1
                try:
                    self.mapper.verify_legacy_record(legacy_record, idx)
                    folio_rec, legacy_id = self.mapper.do_map(
                        legacy_record, f"row # {idx}", FOLIONamespaces.holdings
                    )
                    self.post_process_holding(folio_rec, legacy_id)
                except TransformationProcessError as process_error:
                    self.mapper.handle_transformation_process_error(idx, process_error)
                except TransformationRecordFailedError as error:
                    self.mapper.handle_transformation_record_failed_error(idx, error)
                except Exception as excepion:
                    self.mapper.handle_generic_exception(idx, excepion)
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Number of Legacy items in file")
                )
                if idx > 1 and idx % 10000 == 0:
                    elapsed = idx / (time.time() - start)
                    elapsed_formatted = "{0:.4g}".format(elapsed)
                    logging.info(f"{idx:,} records processed. Recs/sec: {elapsed_formatted} ")
            self.total_records = records_processed
            logging.info(
                f"Done processing {file_def.file_name} containing {self.total_records:,} records. "
                f"Total records processed: {self.total_records:,}"
            )

    def post_process_holding(self, folio_rec: dict, legacy_id: str):
        HoldingsHelper.handle_notes(folio_rec)
        HoldingsHelper.remove_empty_holdings_statements(folio_rec)

        if not folio_rec.get("holdingsTypeId", ""):
            folio_rec["holdingsTypeId"] = self.fallback_holdings_type["id"]

        folio_rec["sourceId"] = self.holdings_sources.get("FOLIO")

        holdings_from_row = []
        all_instance_ids = folio_rec.get("instanceId", [])
        if len(all_instance_ids) == 1:
            # Normal case.
            folio_rec["instanceId"] = folio_rec["instanceId"][0]
            holdings_from_row.append(folio_rec)

        elif len(folio_rec.get("instanceId", [])) > 1:  # Bound-with.
            holdings_from_row.extend(self.create_bound_with_holdings(folio_rec, legacy_id))
        else:
            raise TransformationRecordFailedError(legacy_id, "No instance id in parsed record", "")

        for folio_holding in holdings_from_row:
            self.merge_holding_in(folio_holding, all_instance_ids, legacy_id)
        self.mapper.report_folio_mapping(folio_holding, self.mapper.schema)

    def create_bound_with_holdings(self, folio_holding, legacy_id: str):
        folio_holding["formerIds"] = explode_former_ids(folio_holding)
        return list(
            self.mapper.create_bound_with_holdings(
                folio_holding,
                folio_holding["instanceId"],
                self.task_config.holdings_type_uuid_for_boundwiths,
            )
        )

    def merge_holding_in(
        self, incoming_holding: dict, instance_ids: list[str], legacy_item_id: str
    ) -> None:
        """Determines what newly generated holdingsrecords are to be merged with
        previously created ones. When that is done, it generates the correct boundwith
        parts needed.

        Args:
            incoming_holding (dict): The newly created FOLIO Holding
            instance_ids (list): the instance IDs tied to the current item
            legacy_item_id (str): Id of the Item the holding was generated from
        """
        if len(instance_ids) > 1:
            # Is boundwith
            bw_key = f"bw_{incoming_holding['instanceId']}_{'_'.join(sorted(instance_ids))}"
            if bw_key not in self.bound_with_keys:
                self.bound_with_keys.add(bw_key)
                self.holdings[bw_key] = incoming_holding
                self.mapper.create_and_write_boundwith_part(legacy_item_id, incoming_holding["id"])
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Unique BW Holdings created from Items")
                )
            else:
                self.merge_holding(bw_key, incoming_holding)
                self.mapper.create_and_write_boundwith_part(
                    legacy_item_id, self.holdings[bw_key]["id"]
                )
                self.holdings_id_map[legacy_item_id] = self.mapper.get_id_map_tuple(
                    legacy_item_id, self.holdings[bw_key], self.object_type
                )
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("BW Items found tied to previously created BW Holdings")
                )
        else:
            # Regular holding. Merge according to criteria
            new_holding_key = HoldingsHelper.to_key(
                incoming_holding,
                self.task_config.holdings_merge_criteria,
                self.mapper.migration_report,
                self.task_config.holdings_type_uuid_for_boundwiths,
            )
            if self.holdings.get(new_holding_key, None):
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Holdings already created from Item")
                )
                self.merge_holding(new_holding_key, incoming_holding)
            else:
                self.mapper.migration_report.add_general_statistics(
                    i18n.t("Unique Holdings created from Items")
                )
                self.holdings[new_holding_key] = incoming_holding

    def merge_holding(self, holdings_key: str, new_holdings_record: dict):
        self.holdings[holdings_key] = HoldingsHelper.merge_holding(
            self.holdings[holdings_key], new_holdings_record
        )

    def get_holdings_sources(self):
        res = {}
        holdings_sources = list(
            self.mapper.folio_client.folio_get_all("/holdings-sources", "holdingsRecordsSources")
        )
        logging.info("Fetched %s holdingsRecordsSources from tenant", len(holdings_sources))
        res = {n["name"].upper(): n["id"] for n in holdings_sources}
        if "FOLIO" not in res:
            raise TransformationProcessError("", "No holdings source with name FOLIO in tenant")
        if "MARC" not in res:
            raise TransformationProcessError("", "No holdings source with name MARC in tenant")
        logging.info(json.dumps(res, indent=4))
        return res


def explode_former_ids(folio_holding: dict):
    temp_ids = []
    for former_id in folio_holding.get("formerIds", []):
        if former_id.startswith("[") and former_id.endswith("]") and "," in former_id:
            ids = list(
                former_id[1:-1].replace('"', "").replace(" ", "").replace("'", "").split(",")
            )
            temp_ids.extend(ids)
        else:
            temp_ids.append(former_id)
    return temp_ids



import logging
from datetime import datetime
from zoneinfo import ZoneInfo

from dateutil import tz
from dateutil.parser import parse

from folio_migration_tools.migration_report import MigrationReport

utc = ZoneInfo("UTC")


class LegacyLoan(object):
    def __init__(
        self,
        legacy_loan_dict,
        fallback_service_point_id: str,
        migration_report: MigrationReport,
        tenant_timezone=utc,
        row=0,
    ):
        self.migration_report: MigrationReport = migration_report
        # validate
        correct_headers = [
            "item_barcode",
            "patron_barcode",
            "due_date",
            "out_date",
            "renewal_count",
            "next_item_status",
            "service_point_id",
        ]
        optional_headers = ["service_point_id"]
        legal_statuses = [
            "",
            "Aged to lost",
            "Checked out",
            "Claimed returned",
            "Declared lost",
            "Lost and paid",
        ]

        self.tenant_timezone = tenant_timezone
        self.errors = []
        for prop in correct_headers:
            if prop not in legacy_loan_dict and prop not in optional_headers:
                self.errors.append(("Missing properties in legacy data", prop))
            if (
                prop != "next_item_status"
                and not legacy_loan_dict.get(prop, "").strip()
                and prop not in optional_headers
            ):
                self.errors.append(("Empty properties in legacy data", prop))
        try:
            temp_date_due: datetime = parse(legacy_loan_dict["due_date"])
            if temp_date_due.tzinfo != tz.UTC:
                temp_date_due = temp_date_due.replace(tzinfo=self.tenant_timezone)
                self.report(
                    f"Provided due_date is not UTC, "
                    f"setting tzinfo to tenant timezone ({self.tenant_timezone})"
                )
            if temp_date_due.hour == 0 and temp_date_due.minute == 0:
                temp_date_due = temp_date_due.replace(hour=23, minute=59)
                self.report(
                    "Hour and minute not specified for due date. "
                    "Assuming end of local calendar day (23:59)..."
                )
        except Exception as ee:
            logging.error(ee)
            self.errors.append(("Parse date failure. Setting UTC NOW", "due_date"))
            temp_date_due = datetime.now(ZoneInfo("UTC"))
        try:
            temp_date_out: datetime = parse(legacy_loan_dict["out_date"])
            if temp_date_out.tzinfo != tz.UTC:
                temp_date_out = temp_date_out.replace(tzinfo=self.tenant_timezone)
                self.report(
                    f"Provided out_date is not UTC, "
                    f"setting tzinfo to tenant timezone ({self.tenant_timezone})"
                )
        except Exception:
            temp_date_out = datetime.now(
                ZoneInfo("UTC")
            )  # TODO: Consider moving this assignment block above the temp_date_due
            self.errors.append(("Parse date failure. Setting UTC NOW", "out_date"))

        # good to go, set properties
        self.item_barcode: str = legacy_loan_dict["item_barcode"].strip()
        self.patron_barcode: str = legacy_loan_dict["patron_barcode"].strip()
        self.due_date: datetime = temp_date_due
        self.out_date: datetime = temp_date_out
        self.correct_for_1_day_loans()
        self.make_utc()
        self.renewal_count = int(legacy_loan_dict["renewal_count"])
        self.next_item_status = legacy_loan_dict.get("next_item_status", "").strip()
        if self.next_item_status not in legal_statuses:
            self.errors.append(("Not an allowed status", self.next_item_status))
        self.service_point_id = (
            legacy_loan_dict["service_point_id"]
            if legacy_loan_dict.get("service_point_id", "")
            else fallback_service_point_id
        )

    def correct_for_1_day_loans(self):
        try:
            if self.due_date.date() <= self.out_date.date():
                if self.due_date.hour == 0:
                    self.due_date = self.due_date.replace(hour=23, minute=59)
                if self.out_date.hour == 0:
                    self.out_date = self.out_date.replace(hour=0, minute=1)
            if self.due_date <= self.out_date:
                raise ValueError("Due date is before out date")
        except Exception:
            self.errors.append(("Time alignment issues", "both dates"))

    def to_dict(self):
        return {
            "item_barcode": self.item_barcode,
            "patron_barcode": self.patron_barcode,
            "due_date": self.due_date.isoformat(),
            "out_date": self.out_date.isoformat(),
            "renewal_count": self.renewal_count,
            "next_item_status": self.next_item_status,
        }

    def make_utc(self):
        try:
            if self.tenant_timezone != ZoneInfo("UTC"):
                self.due_date = self.due_date.astimezone(ZoneInfo("UTC"))
                self.out_date = self.out_date.astimezone(ZoneInfo("UTC"))
        except Exception:
            self.errors.append(("UTC correction issues", "both dates"))

    def report(self, what_to_report: str):
        self.migration_report.add("Details", what_to_report)



import json
import logging
import os
import re
import sys
import urllib.parse
import uuid
import i18n

import httpx
from folio_uuid.folio_uuid import FOLIONamespaces
from folioclient import FolioClient
from httpx import HTTPError

from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.notes_mapper import NotesMapper
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class CompositeOrderMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        library_configuration: LibraryConfiguration,
        composite_order_map: dict,
        organizations_id_map: dict,
        instance_id_map: dict,
        acquisition_method_map,
        payment_status_map,
        receipt_status_map,
        workflow_status_map,
        location_map,
        funds_map,
        funds_expense_class_map=None,
    ):
        # Get organization schema
        self.composite_order_schema = CompositeOrderMapper.get_latest_acq_schemas_from_github(
            "folio-org", "mod-orders", "mod-orders", "composite_purchase_order"
        )

        super().__init__(
            folio_client,
            self.composite_order_schema,
            composite_order_map,
            None,
            FOLIONamespaces.orders,
            library_configuration,
        )
        logging.info("Loading Instance ID map...")
        self.instance_id_map = instance_id_map
        self.organizations_id_map = organizations_id_map
        self.folio_organization_cache = {}

        self.acquisitions_methods_mapping = RefDataMapping(
            self.folio_client,
            "/orders/acquisition-methods",
            "acquisitionMethods",
            acquisition_method_map,
            "value",
            "AcquisitionMethodMapping",
        )
        logging.info("Init done")
        self.location_mapping = RefDataMapping(
            self.folio_client,
            "/locations",
            "locations",
            location_map,
            "code",
            "OrderLineLocationMapping",
        )

        self.folio_client: FolioClient = folio_client
        self.notes_mapper: NotesMapper = NotesMapper(
            library_configuration,
            self.folio_client,
            composite_order_map,
            FOLIONamespaces.note,
            True,
        )
        self.notes_mapper.migration_report = self.migration_report

    def get_prop(self, legacy_order, folio_prop_name: str, index_or_id, schema_default_value):
        if folio_prop_name.endswith(".acquisitionMethod"):
            return self.get_mapped_ref_data_value(
                self.acquisitions_methods_mapping,
                legacy_order,
                folio_prop_name,
                index_or_id,
                False,
            )

        elif re.compile(r"compositePoLines\[(\d+)\]\.id").fullmatch(folio_prop_name):
            return str(uuid.uuid4())

        elif re.compile(r"notes\[\d+\]\.").match(folio_prop_name):
            return ""

        if folio_prop_name.endswith(".locationId"):
            return self.get_mapped_ref_data_value(
                self.location_mapping,
                legacy_order,
                folio_prop_name,
                index_or_id,
                False,
            )

        mapped_value = super().get_prop(
            legacy_order, folio_prop_name, index_or_id, schema_default_value
        )

        return mapped_value

    @staticmethod
    def get_latest_acq_schemas_from_github(owner, repo, module, object):
        """
        Given a repository owner, a repository, a module name and the name
        of a FOLIO acquisition object, returns a schema for that object that
        also includes the schemas of any other referenced acq objects.

        Args:
            owner (_type_): _description_
            repo (_type_): _description_
            module (_type_): _description_
            object (_type_): _description_

        Returns:
            _type_: _description_
        """
        try:
            # Authenticate when calling GitHub, using an API key stored in .env
            github_headers = {
                "content-type": "application/json",
                "User-Agent": "FOLIO Migration Tools (https://github.com/FOLIO-FSE/folio_migration_tools/)",  # noqa:E501,B950
            }

            if os.environ.get("GITHUB_TOKEN"):
                logging.info("Using GITHB_TOKEN environment variable for Gihub API Access")
                github_headers["authorization"] = f"token {os.environ.get('GITHUB_TOKEN')}"

            # Start talkign to GitHub...
            github_path = "https://raw.githubusercontent.com"
            submodules = CompositeOrderMapper.get_submodules_of_latest_release(
                owner, repo, github_headers
            )

            # Get the sha's of submodules acq-models and raml_utils
            acq_models_sha = next(
                (item["sha"] for item in submodules if item["path"] == "acq-models")
            )

            # # TODO Maybe - fetch raml_utils schemas if deemed necessary
            # raml_utils_sha = next((item["sha"] for item in submodules
            # if item["path"] == "raml-utils"))

            acq_models_path = (
                f"{github_path}/{owner}/acq-models/{acq_models_sha}/{module}/schemas/"
            )

            req = httpx.get(
                f"{acq_models_path}/{object}.json",
                headers=github_headers,
                follow_redirects=True,
                timeout=None,
            )
            req.raise_for_status()

            object_schema = json.loads(req.text)

            return CompositeOrderMapper.build_extended_object(
                object_schema, acq_models_path, github_headers
            )
        except httpx.HTTPError as http_error:
            logging.critical(f"Halting! \t{http_error}")
            sys.exit(2)

        except json.decoder.JSONDecodeError as json_error:
            logging.critical(json_error)
            sys.exit(2)

    @staticmethod
    def get_submodules_of_latest_release(owner, repo, github_headers):
        """
        Given a repository owner and a repository, identifies the latest
        release of the repository and returns the submodules associated with
        this release.

        Args:
            owner (_type_): _description_
            repo (_type_): _description_
            github_headers (_type_): _description_

        Returns:
            _type_: _description_


        """

        github_path = "https://api.github.com/repos"

        # Get metadata for the latest release
        latest_release_path = f"{github_path}/{owner}/{repo}/releases/latest"
        req = httpx.get(
            f"{latest_release_path}", headers=github_headers, follow_redirects=True, timeout=None
        )
        req.raise_for_status()
        latest_release = json.loads(req.text)

        # Get the tag assigned to the latest release
        release_tag = latest_release["tag_name"]
        logging.info(f"Using schemas from latest {repo} release: {release_tag}")

        # Get the tree for the latest release
        tree_path = f"{github_path}/{owner}/{repo}/git/trees/{release_tag}"
        req = httpx.get(tree_path, headers=github_headers, follow_redirects=True, timeout=None)
        req.raise_for_status()
        release_tree = json.loads(req.text)

        # Loop through the tree to get the sha of the folder with path "ramls"
        ramls_sha = next((item["sha"] for item in release_tree["tree"] if item["path"] == "ramls"))

        # Get the tree for the ramls folder
        ramls_path = f"{github_path}/{owner}/{repo}/git/trees/{ramls_sha}"
        req = httpx.get(ramls_path, headers=github_headers, follow_redirects=True, timeout=None)
        req.raise_for_status()
        ramls_tree = json.loads(req.text)

        # Loop through the tree to get the sha of submodules
        submodules = [item for item in ramls_tree["tree"] if item["mode"] == "160000"]

        return submodules

    @staticmethod
    def build_extended_object(object_schema, submodule_path, github_headers):
        """
        Takes an object schema (for example an organization) and the path to a
        submodule repository and returns the same schema with the full schemas
        of subordinate objects (for example aliases).

        Args:
            object_schema (_type_): _description_
            submodule_path (_type_): _description_
            github_headers (_type_): _description_

        Returns:
            _type_: _description_
        """

        supported_types = [
            "string",
            "boolean",
            "number",
            "integer",
            "text",
            "object",
            "array",
        ]

        try:
            if (
                "properties" not in object_schema
                and "$ref" in object_schema
                and object_schema["type"] == "object"
            ):
                submodule_path["properties"] = CompositeOrderMapper.inject_schema_by_ref(
                    submodule_path, github_headers, object_schema
                ).get("properties")

            for property_name_level1, property_level1 in object_schema.get(
                "properties", {}
            ).items():
                # Report and discard unhandled properties
                if property_level1.get("type") not in supported_types:
                    logging.info(f"Property not yet supported: {property_name_level1}")
                    property_level1["type"] = "Deprecated"

                # Handle object properties
                elif property_level1.get("type") == "object" and property_level1.get("$ref"):
                    logging.info("Fecthing referenced schema for object %s", property_name_level1)
                    actual_path = urllib.parse.urljoin(
                        f"{submodule_path}", object_schema.get("$ref", "")
                    )

                    p1 = CompositeOrderMapper.inject_schema_by_ref(
                        actual_path, github_headers, property_level1
                    )

                    p2 = CompositeOrderMapper.build_extended_object(
                        p1, actual_path, github_headers
                    )
                    object_schema["properties"][property_name_level1] = p2

                # Handle arrays of items properties
                elif property_level1.get("type") == "array" and property_level1.get("items").get(
                    "$ref"
                ):
                    logging.info(
                        "Fetching referenced schema for array object %s", property_name_level1
                    )
                    actual_path = urllib.parse.urljoin(
                        f"{submodule_path}", object_schema.get("$ref", "")
                    )

                    p1 = CompositeOrderMapper.inject_items_schema_by_ref(
                        actual_path, github_headers, property_level1
                    )
                    p2 = CompositeOrderMapper.build_extended_object(
                        p1, actual_path, github_headers
                    )
                    property_level1["items"] = p2
                elif property_level1.get("type") == "string" and property_level1.get("$ref"):
                    logging.info("Fetching referenced schema for object %s", property_name_level1)
                    actual_path = urllib.parse.urljoin(
                        f"{submodule_path}", object_schema.get("$ref", "")
                    )
                    p1 = CompositeOrderMapper.inject_schema_by_ref(
                        actual_path, github_headers, property_level1
                    )
                    object_schema["properties"][property_name_level1] = p1

            return object_schema

        except HTTPError as he:
            logging.error(he)

    @staticmethod
    def inject_schema_by_ref(submodule_path, github_headers, property: dict):
        base_raml = "https://raw.githubusercontent.com/folio-org/raml/master/"
        try:
            u1 = urllib.parse.urlparse(submodule_path)
            schema_url = urllib.parse.urljoin(u1.geturl(), property["$ref"])
            if schema_url.endswith("tags.schema"):
                schema_url = f"{base_raml}schemas/tags.schema"
            if schema_url.endswith("metadata.schema"):
                schema_url = f"{base_raml}schemas/metadata.schema"

            req = httpx.get(
                schema_url, headers=github_headers, follow_redirects=True, timeout=None
            )
            req.raise_for_status()
            return dict(property, **json.loads(req.text))
        except Exception as ee:
            logging.error(ee)
            return {}

    @staticmethod
    def inject_items_schema_by_ref(submodule_path, github_headers, property: dict):
        try:
            u1 = urllib.parse.urlparse(submodule_path)
            schema_url = urllib.parse.urljoin(u1.geturl(), property["items"]["$ref"])
            req = httpx.get(
                schema_url, headers=github_headers, follow_redirects=True, timeout=None
            )
            req.raise_for_status()
            return dict(property["items"], **json.loads(req.text))
        except Exception as ee:
            logging.error(ee)
            return {}

    def perform_additional_mapping(self, index_or_id, composite_order):
        # Get organization UUID from FOLIO
        composite_order["vendor"] = self.get_folio_organization_uuid(
            index_or_id, composite_order.get("vendor")
        )

        # Replace legacy bib ID with instance UUID from map
        bib_id = composite_order["compositePoLines"][0].pop("instanceId", "")
        if matching_instance := self.get_folio_instance_uuid(index_or_id, bib_id):
            composite_order["compositePoLines"][0]["instanceId"] = matching_instance

        return composite_order

    def get_matching_record_from_folio(
        self,
        index_or_id,
        cache: dict,
        path: str,
        match_property: str,
        match_value: str,
        result_type: str,
    ):
        if match_value in cache:
            return cache[match_value]
        else:
            query = f'?query=({match_property}=="{match_value}")'
            if matching_record := next(
                self.folio_client.folio_get_all(path, result_type, query), None
            ):
                cache[match_value] = matching_record
                return matching_record

    def get_folio_organization_uuid(self, index_or_id, org_code):
        if self.organizations_id_map:
            self.migration_report.add(
                "PurchaseOrderVendorLinking",
                i18n.t("Organizations linked using organizations_id_map"),
            )
            if matching_org := self.organizations_id_map.get(org_code):
                return matching_org[1]

        if matching_org := self.get_matching_record_from_folio(
            index_or_id,
            self.folio_organization_cache,
            "/organizations-storage/organizations",
            "code",
            org_code,
            "organizations",
        ):
            self.migration_report.add(
                "PurchaseOrderVendorLinking",
                i18n.t("Organizations not in ID map, linked using FOLIO lookup"),
            )
            return matching_org["id"]

        else:
            self.migration_report.add(
                "PurchaseOrderVendorLinking",
                i18n.t("RECORD FAILED Organization identifier not in ID map/FOLIO"),
            )
            raise TransformationRecordFailedError(
                index_or_id,
                "No matching Organization for organization identifier",
                org_code,
            )

    def get_folio_instance_uuid(self, index_or_id, bib_id):
        if matching_instance := self.instance_id_map.get(bib_id):
            self.migration_report.add(
                "PurchaseOrderInstanceLinking",
                i18n.t("Instances linked using instances_id_map"),
            )
            return matching_instance[1]
        else:
            self.migration_report.add(
                "PurchaseOrderInstanceLinking",
                i18n.t("Bib identifier not in instances_id_map, no instance linked"),
            )
            Helper.log_data_issue(
                index_or_id,
                "No matching instance for bib identifier",
                bib_id,
            )



import uuid
from typing import Dict
from typing import List
from typing import Tuple

from folio_uuid.folio_namespaces import FOLIONamespaces
from folio_uuid.folio_uuid import FolioUUID
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationProcessError


class LegacyReserve(object):
    def __init__(self, legacy_request_dict: Dict, folio_client: FolioClient, row: int = 0):
        # validate
        correct_headers = ["legacy_identifier", "item_barcode"]
        for h in correct_headers:
            if h not in legacy_request_dict:
                raise TransformationProcessError(
                    int,
                    "Missing header in file. The following are required:",
                    ",".join(correct_headers),
                )
        self.errors: List[Tuple[str, str]] = [
            ("Missing properties in legacy data", prop)
            for prop in correct_headers
            if prop not in legacy_request_dict
        ]
        self.id = str(uuid.uuid4())
        self.item_barcode: str = legacy_request_dict["item_barcode"].strip()
        if not self.item_barcode:
            self.errors.append(("Missing data.", "item_barcode"))
        self.legacy_identifier: str = legacy_request_dict["legacy_identifier"].strip()
        if not self.legacy_identifier:
            self.errors.append(("Missing data.", "legacy_identifier"))
        self.course_listing_id: str = str(
            FolioUUID(
                folio_client.okapi_url,
                FOLIONamespaces.course_listing,
                legacy_request_dict["legacy_identifier"],
            )
        )

    def to_dict(self):
        return {
            "courseListingId": self.course_listing_id,
            "copiedItem": {"barcode": self.item_barcode},
            "id": self.id,
        }



import ast
import i18n

from folio_uuid.folio_uuid import FOLIONamespaces
from folioclient import FolioClient

from folio_migration_tools.custom_exceptions import TransformationRecordFailedError
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.mapping_file_transformation.ref_data_mapping import (
    RefDataMapping,
)


class HoldingsMapper(MappingFileMapperBase):
    def __init__(
        self,
        folio_client: FolioClient,
        holdings_map,
        location_map,
        call_number_type_map,
        instance_id_map,
        library_configuration: LibraryConfiguration,
        statistical_codes_map=None,
    ):
        holdings_schema = folio_client.get_holdings_schema()
        self.instance_id_map = instance_id_map
        super().__init__(
            folio_client,
            holdings_schema,
            holdings_map,
            statistical_codes_map,
            FOLIONamespaces.holdings,
            library_configuration,
        )
        self.holdings_map = holdings_map

        self.location_mapping = RefDataMapping(
            self.folio_client,
            "/locations",
            "locations",
            location_map,
            "code",
            "LocationMapping",
        )
        if call_number_type_map:
            self.call_number_mapping = RefDataMapping(
                self.folio_client,
                "/call-number-types",
                "callNumberTypes",
                call_number_type_map,
                "name",
                "CallNumberTypeMapping",
            )

    def perform_additional_mappings(self, folio_rec, file_def):
        self.handle_suppression(folio_rec, file_def)

    def handle_suppression(self, folio_record, file_def: FileDefinition):
        folio_record["discoverySuppress"] = file_def.discovery_suppressed
        self.migration_report.add(
            "Suppression",
            i18n.t("Suppressed from discovery") + f' = {folio_record["discoverySuppress"]}',
        )

    def get_prop(self, legacy_item, folio_prop_name, index_or_id, schema_default_value):
        if folio_prop_name == "permanentLocationId":
            return self.get_location_id(legacy_item, index_or_id, folio_prop_name)
        elif folio_prop_name == "callNumberTypeId":
            return self.get_call_number_type_id(legacy_item, folio_prop_name, index_or_id)
        elif folio_prop_name.startswith("statisticalCodeIds"):
            return self.get_statistical_code(legacy_item, folio_prop_name, index_or_id)

        mapped_value = super().get_prop(
            legacy_item, folio_prop_name, index_or_id, schema_default_value
        )
        if folio_prop_name == "callNumber":
            return self.get_call_number(mapped_value)
        elif folio_prop_name == "instanceId":
            return self.get_instance_ids(mapped_value, index_or_id)
        elif mapped_value:
            return mapped_value
        else:
            self.migration_report.add("UnmappedProperties", f"{folio_prop_name}")
            return ""

    def get_call_number(self, legacy_value):
        if legacy_value.startswith("[") and len(legacy_value.split(",")) > 1:
            self.migration_report.add_general_statistics(
                i18n.t("Bound-with items callnumber identified")
            )
            self.migration_report.add(
                "BoundWithMappings",
                (f"Number of bib-level callnumbers in record: {len(legacy_value.split(','))}"),
            )
        return legacy_value

    def get_location_id(
        self, legacy_item: dict, id_or_index, folio_prop_name, prevent_default=False
    ):
        return self.get_mapped_ref_data_value(
            self.location_mapping,
            legacy_item,
            id_or_index,
            folio_prop_name,
            prevent_default,
        )

    def get_call_number_type_id(self, legacy_item, folio_prop_name: str, id_or_index):
        if self.call_number_mapping:
            return self.get_mapped_ref_data_value(
                self.call_number_mapping, legacy_item, id_or_index, folio_prop_name
            )
        self.migration_report.add("CallNumberTypeMapping", i18n.t("No Call Number Type Mapping"))
        return ""

    def get_instance_ids(self, legacy_value: str, index_or_id: str):
        # Returns a list of Id:s
        return_ids = []
        legacy_bib_ids = self.get_legacy_bib_ids(legacy_value, index_or_id)
        self.migration_report.add(
            "BoundWithMappings",
            i18n.t("Number of bib records referenced in item") + f": {len(legacy_bib_ids)}",
        )
        for legacy_instance_id in legacy_bib_ids:
            new_legacy_value = (
                f".{legacy_instance_id}"
                if legacy_instance_id.startswith("b")
                else legacy_instance_id
            )
            if (
                new_legacy_value not in self.instance_id_map
                and legacy_instance_id not in self.instance_id_map
            ):
                self.migration_report.add_general_statistics(
                    i18n.t("Records not matched to Instances")
                )
                s = "Bib id not in instance id map."
                raise TransformationRecordFailedError(index_or_id, s, new_legacy_value)
            else:
                self.migration_report.add_general_statistics(
                    i18n.t("Records matched to Instances")
                )
                entry = self.instance_id_map.get(new_legacy_value, "") or self.instance_id_map.get(
                    legacy_instance_id
                )
                return_ids.append(entry[1])
        if any(return_ids):
            return return_ids
        else:
            raise TransformationRecordFailedError(
                index_or_id, "No instance id mapped from", legacy_value
            )

    def get_legacy_bib_ids(self, legacy_value: str, index_or_id: str) -> list[str]:
        if not legacy_value.startswith("["):
            return [legacy_value]
        try:
            new_legacy_values = ast.literal_eval(legacy_value)
            new_value_len = len(new_legacy_values)
            if new_value_len > 1:
                self.migration_report.add_general_statistics(
                    i18n.t("Bound-with items identified by bib id")
                )
                self.migration_report.add(
                    "GeneralStatistics",
                    i18n.t("Bib ids referenced in bound-with items"),
                    new_value_len,
                )
            return new_legacy_values
        except Exception as error:
            raise TransformationRecordFailedError(
                index_or_id,
                f"Instance ID could not get parsed to array of strings {error}",
                legacy_value,
            ) from error



import copy
import csv
import json
import logging
import sys
import time
import traceback
import i18n
from datetime import datetime
from datetime import timedelta
from typing import Optional
from urllib.error import HTTPError
from zoneinfo import ZoneInfo

import httpx
from dateutil import parser as du_parser
from folio_uuid.folio_namespaces import FOLIONamespaces

from folio_migration_tools.circulation_helper import CirculationHelper
from folio_migration_tools.helper import Helper
from folio_migration_tools.library_configuration import FileDefinition
from folio_migration_tools.library_configuration import FolioRelease
from folio_migration_tools.library_configuration import LibraryConfiguration
from folio_migration_tools.mapping_file_transformation.mapping_file_mapper_base import (
    MappingFileMapperBase,
)
from folio_migration_tools.migration_report import MigrationReport
from folio_migration_tools.migration_tasks.migration_task_base import MigrationTaskBase
from folio_migration_tools.task_configuration import AbstractTaskConfiguration
from folio_migration_tools.transaction_migration.legacy_loan import LegacyLoan
from folio_migration_tools.transaction_migration.transaction_result import (
    TransactionResult,
)


class LoansMigrator(MigrationTaskBase):
    class TaskConfiguration(AbstractTaskConfiguration):
        name: str
        migration_task_type: str
        open_loans_files: list[FileDefinition]
        fallback_service_point_id: str
        starting_row: Optional[int] = 1
        item_files: Optional[list[FileDefinition]] = []
        patron_files: Optional[list[FileDefinition]] = []

    @staticmethod
    def get_object_type() -> FOLIONamespaces:
        return FOLIONamespaces.loans

    def __init__(
        self,
        task_configuration: TaskConfiguration,
        library_config: LibraryConfiguration,
    ):
        csv.register_dialect("tsv", delimiter="\t")
        self.patron_item_combos: set = set()
        self.t0 = time.time()
        self.num_duplicate_loans = 0
        self.skipped_since_already_added = 0
        self.processed_items: set = set()
        self.failed: dict = {}
        self.failed_and_not_dupe: dict = {}
        self.migration_report = MigrationReport()
        self.valid_legacy_loans = []
        super().__init__(library_config, task_configuration)
        self.circulation_helper = CirculationHelper(
            self.folio_client,
            task_configuration.fallback_service_point_id,
            self.migration_report,
        )
        logging.info("Check that SMTP is disabled before migrating loans")
        self.check_smtp_config()
        logging.info("Proceeding with loans migration")
        logging.info("Attempting to retrieve tenant timezone configuration...")
        my_path = "/configurations/entries?query=(module==ORG%20and%20configName==localeSettings)"
        try:
            self.tenant_timezone_str = json.loads(
                self.folio_client.folio_get_single_object(my_path)["configs"][0]["value"]
            )["timezone"]
            logging.info("Tenant timezone is: %s", self.tenant_timezone_str)
        except Exception:
            logging.info('Tenant locale settings not available. Using "UTC".')
            self.tenant_timezone_str = "UTC"
        self.tenant_timezone = ZoneInfo(self.tenant_timezone_str)
        self.semi_valid_legacy_loans = []
        for file_def in task_configuration.open_loans_files:
            loans_file_path = self.folder_structure.legacy_records_folder / file_def.file_name
            with open(loans_file_path, "r", encoding="utf-8") as loans_file:
                total_rows, empty_rows, reader = MappingFileMapperBase._get_delimited_file_reader(
                    loans_file, loans_file_path
                )
                logging.info("Source data file contains %d rows", total_rows)
                logging.info("Source data file contains %d empty rows", empty_rows)
                self.migration_report.set(
                    "GeneralStatistics",
                    f"Total rows in {loans_file_path.name}",
                    total_rows,
                )
                self.migration_report.set(
                    "GeneralStatistics",
                    f"Empty rows in {loans_file_path.name}",
                    empty_rows,
                )
                self.semi_valid_legacy_loans.extend(
                    self.load_and_validate_legacy_loans(
                        reader,
                        file_def.service_point_id or task_configuration.fallback_service_point_id,
                    )
                )

                logging.info(
                    "Loaded and validated %s loans in file from %s",
                    len(self.semi_valid_legacy_loans),
                    file_def.file_name,
                )
        logging.info("Loaded and validated %s loans in total", len(self.semi_valid_legacy_loans))
        if any(self.task_configuration.item_files) or any(self.task_configuration.patron_files):
            self.valid_legacy_loans = list(self.check_barcodes())
            logging.info(
                "Loaded and validated %s loans against barcodes",
                len(self.valid_legacy_loans),
            )
        else:
            logging.info(
                "No item or user files supplied. Not validating against"
                "previously migrated objects"
            )
            self.valid_legacy_loans = self.semi_valid_legacy_loans
        logging.info("Starting row number is %s", task_configuration.starting_row)
        logging.info("Init completed")

    def check_smtp_config(self):
        if self.library_configuration.folio_release.lower() == FolioRelease.morning_glory:
            logging.warn("DEPRECATED: Morning Glory support will be removed in a future release!")
            okapi_config_base_path = "/configurations/entries"
            okapi_config_query = "(module=={}%20and%20configName=={}%20and%20code=={})"
            okapi_config_limit = 1000
            okapi_config_module = "SMTP_SERVER"
            okapi_config_name = "smtp"
            okapi_config_code = "EMAIL_SMTP_HOST_DISABLED"
            smtp_config_path = (
                okapi_config_base_path
                + "?"
                + str(okapi_config_limit)
                + "&query="
                + okapi_config_query.format(
                    okapi_config_module, okapi_config_name, okapi_config_code
                )
            )
            smtp_config_disabled = self.folio_client.folio_get_single_object(smtp_config_path)[
                "configs"
            ]
        else:
            try:
                smtp_config = self.folio_client.folio_get_single_object("/smtp-configuration")[
                    "smtpConfigurations"
                ][0]
                smtp_config_disabled = "disabled" in smtp_config["host"].lower()
            except IndexError:
                smtp_config_disabled = True
        print_smtp_warning()
        if not smtp_config_disabled:
            logging.warn("SMTP connection not disabled...")
            for i in range(10, 0, -1):
                sys.stdout.write("Pausing for {:02d} seconds. Press Ctrl+C to exit...\r".format(i))
                time.sleep(1)
        else:
            logging.info("SMTP connection is disabled...")

    def do_work(self):
        with httpx.Client(
            timeout=None, headers=self.folio_client.okapi_headers
        ) as self.http_client:
            logging.info("Starting")
            starting_index = (
                self.task_configuration.starting_row - 1
                if self.task_configuration.starting_row > 0
                else 0
            )
            if self.task_configuration.starting_row > 1:
                logging.info(f"Skipping {(starting_index)} records")
            for num_loans, legacy_loan in enumerate(
                self.valid_legacy_loans[starting_index:], start=1
            ):
                t0_migration = time.time()
                self.migration_report.add_general_statistics(
                    i18n.t("Processed pre-validated loans")
                )
                try:
                    self.checkout_single_loan(legacy_loan)
                except Exception as ee:
                    logging.exception(
                        f"Error in row {num_loans}  Item barcode: {legacy_loan.item_barcode} "
                        f"Patron barcode: {legacy_loan.patron_barcode} {ee}"
                    )
                if num_loans % 25 == 0:
                    logging.info(f"{timings(self.t0, t0_migration, num_loans)} {num_loans}")

    def checkout_single_loan(self, legacy_loan: LegacyLoan):
        """Checks a legacy loan out. Retries once if it fails.

        Args:
            legacy_loan (LegacyLoan): The Legacy loan
        """
        res_checkout = self.circulation_helper.check_out_by_barcode(legacy_loan)

        if res_checkout.was_successful:
            self.migration_report.add("Details", i18n.t("Checked out on first try"))
            self.migration_report.add_general_statistics(i18n.t("Successfully checked out"))
            self.set_renewal_count(legacy_loan, res_checkout)
            self.set_new_status(legacy_loan, res_checkout)
        elif res_checkout.should_be_retried:
            res_checkout2 = self.handle_checkout_failure(legacy_loan, res_checkout)
            if res_checkout2.was_successful and res_checkout2.folio_loan:
                self.migration_report.add("Details", i18n.t("Checked out on second try"))
                self.migration_report.add_general_statistics(i18n.t("Successfully checked out"))
                logging.info("Checked out on second try")
                self.set_renewal_count(legacy_loan, res_checkout2)
                self.set_new_status(legacy_loan, res_checkout2)
            elif legacy_loan.item_barcode not in self.failed:
                if res_checkout2.error_message == "Aged to lost and checked out":
                    self.migration_report.add(
                        "Details",
                        i18n.t("Second failure") + f": {res_checkout2.migration_report_message}",
                    )
                    logging.error(
                        f"{res_checkout2.error_message}. Item barcode: {legacy_loan.item_barcode}"
                    )
                else:
                    self.failed[legacy_loan.item_barcode] = legacy_loan
                    self.migration_report.add_general_statistics(i18n.t("Failed loans"))
                    Helper.log_data_issue(
                        "", "Loans failing during checkout", json.dumps(legacy_loan.to_dict())
                    )
                    logging.error("Failed on second try: %s", res_checkout2.error_message)
                    self.migration_report.add(
                        "Details",
                        i18n.t("Second failure") + f": {res_checkout2.migration_report_message}",
                    )
        elif not res_checkout.should_be_retried:
            logging.error("Failed first time. No retries: %s", res_checkout.error_message)
            self.migration_report.add_general_statistics(i18n.t("Failed loans"))
            self.migration_report.add(
                "Details",
                i18n.t("Failed 1st time. No retries")
                + f": {res_checkout.migration_report_message}",
            )
            self.failed[legacy_loan.item_barcode] = legacy_loan
            Helper.log_data_issue(
                "", "Loans failing during checkout", json.dumps(legacy_loan.to_dict())
            )

    def set_new_status(self, legacy_loan: LegacyLoan, res_checkout: TransactionResult):
        """Updates checkout loans with their destination statuses

        Args:
            legacy_loan (LegacyLoan): _description_
            res_checkout (TransactionResult): _description_
        """
        # set new statuses
        if legacy_loan.next_item_status == "Declared lost":
            self.declare_lost(res_checkout.folio_loan)
        elif legacy_loan.next_item_status == "Claimed returned":
            self.claim_returned(res_checkout.folio_loan)
        elif legacy_loan.next_item_status not in ["Available", "", "Checked out"]:
            self.set_item_status(legacy_loan)

    def set_renewal_count(self, legacy_loan: LegacyLoan, res_checkout: TransactionResult):
        if legacy_loan.renewal_count > 0:
            self.update_open_loan(res_checkout.folio_loan, legacy_loan)
            self.migration_report.add_general_statistics(i18n.t("Updated renewal count for loan"))

    def wrap_up(self):
        for k, v in self.failed.items():
            self.failed_and_not_dupe[k] = [v.to_dict()]
        print(f"Wrapping up. Unique loans in failed:{len(self.failed_and_not_dupe)}")

        self.write_failed_loans_to_file()

        with open(self.folder_structure.migration_reports_file, "w+") as report_file:
            self.migration_report.write_migration_report(
                i18n.t("Loans migration report"), report_file, self.start_datetime
            )
        self.clean_out_empty_logs()

    def write_failed_loans_to_file(self):
        csv_columns = [
            "due_date",
            "item_barcode",
            "next_item_status",
            "out_date",
            "patron_barcode",
            "renewal_count",
        ]
        with open(self.folder_structure.failed_recs_path, "w+") as failed_loans_file:
            writer = csv.DictWriter(failed_loans_file, fieldnames=csv_columns, dialect="tsv")
            writer.writeheader()
            for _k, failed_loan in self.failed_and_not_dupe.items():
                writer.writerow(failed_loan[0])

    def check_barcodes(self):
        user_barcodes = set()
        item_barcodes = set()
        self.circulation_helper.load_migrated_item_barcodes(
            item_barcodes, self.task_configuration.item_files, self.folder_structure
        )
        self.circulation_helper.load_migrated_user_barcodes(
            user_barcodes, self.task_configuration.patron_files, self.folder_structure
        )
        for loan in self.semi_valid_legacy_loans:
            has_item_barcode = loan.item_barcode in item_barcodes or not any(item_barcodes)
            has_patron_barcode = loan.patron_barcode in user_barcodes or not any(user_barcodes)
            if has_item_barcode and has_patron_barcode:
                self.migration_report.add_general_statistics(
                    i18n.t("Loans verified against migrated user and item")
                )
                yield loan
            else:
                # Add this loan to failed loans for later correction and re-run.
                self.failed[loan.item_barcode] = loan
                self.migration_report.add_general_statistics(i18n.t("Failed loans"))
                self.migration_report.add(
                    "DiscardedLoans",
                    i18n.t("Loans discarded. Had migrated item barcode")
                    + f": {has_item_barcode}. "
                    + i18n.t("Had migrated user barcode")
                    + f": {has_patron_barcode}",
                )
            if not has_item_barcode:
                Helper.log_data_issue(
                    "", "Loan without matched item barcode", json.dumps(loan.to_dict())
                )
            if not has_patron_barcode:
                Helper.log_data_issue(
                    "",
                    "Loan without matched patron barcode",
                    json.dumps(loan.to_dict()),
                )

    def load_and_validate_legacy_loans(self, loans_reader, service_point_id: str) -> list:
        results = []
        num_bad = 0
        logging.info("Validating legacy loans in file...")
        for legacy_loan_count, legacy_loan_dict in enumerate(loans_reader):
            try:
                legacy_loan = LegacyLoan(
                    legacy_loan_dict,
                    service_point_id,
                    self.migration_report,
                    self.tenant_timezone,
                    legacy_loan_count,
                )
                if any(legacy_loan.errors):
                    num_bad += 1
                    self.migration_report.add_general_statistics(
                        i18n.t("Loans failed pre-validation")
                    )
                    self.migration_report.add_general_statistics(i18n.t("Failed loans"))
                    for error in legacy_loan.errors:
                        self.migration_report.add("DiscardedLoans", f"{error[0]} - {error[1]}")
                    # Add this loan to failed loans for later correction and re-run.
                    self.failed[
                        legacy_loan.item_barcode or f"no_barcode_{legacy_loan_count}"
                    ] = legacy_loan
                else:
                    results.append(legacy_loan)
            except ValueError as ve:
                logging.exception(ve)
        logging.info(
            f"Done validating {legacy_loan_count + 1} legacy loans out of which "
            f"{num_bad} where discarded."
        )
        if num_bad / (legacy_loan_count + 1) > 0.5:
            q = num_bad / (legacy_loan_count + 1)
            logging.error("%s percent of loans failed to validate.", (q * 100))
            self.migration_report.log_me()
            logging.critical("Halting...")
            sys.exit(1)
        return results

    def handle_checkout_failure(
        self, legacy_loan, folio_checkout: TransactionResult
    ) -> TransactionResult:
        """Determines what can be done about a previously failed transaction

        Args:
            legacy_loan (_type_): The legacy loan
            folio_checkout (TransactionResult): The results from the prevous transaction

        Returns:
            TransactionResult: A modified TransactionResult based on the result from the
             handling
        """
        folio_checkout.should_be_retried = False
        if folio_checkout.error_message == "5XX":
            return folio_checkout
        if folio_checkout.error_message.startswith(
            "No patron with barcode"
        ) or folio_checkout.error_message.startswith("Patron barcode already detected"):
            return folio_checkout
        elif folio_checkout.error_message.startswith("No item with barcode"):
            return folio_checkout
        elif folio_checkout.error_message.startswith(
            "Cannot check out item that already has an open loan"
        ):
            return folio_checkout
        elif "Aged to lost" in folio_checkout.error_message:
            return self.handle_aged_to_lost_item(legacy_loan)
        elif folio_checkout.error_message == "Declared lost":
            return folio_checkout
        elif folio_checkout.error_message.startswith("Cannot check out to inactive user"):
            return self.checkout_to_inactice_user(legacy_loan)
        else:
            self.migration_report.add(
                "Details",
                i18n.t("Other checkout failure") + f": {folio_checkout.error_message}",
            )
            # First failure. Add to list of failed loans
            if legacy_loan.item_barcode not in self.failed:
                self.failed[legacy_loan.item_barcode] = legacy_loan
            else:
                logging.info(
                    i18n.t("Loan already in failed.")
                    + " "
                    + i18n.t("item barcode")
                    + f": {legacy_loan.item_barcode}"
                    + i18n.t("Patron barcode")
                    + f": {legacy_loan.patron_barcode}",
                )
                self.failed_and_not_dupe[legacy_loan.item_barcode] = [
                    legacy_loan,
                    self.failed[legacy_loan.item_barcode],
                ]
                logging.info(
                    f"Duplicate loans (or failed twice) Item barcode: "
                    f"{legacy_loan.item_barcode} Patron barcode: {legacy_loan.patron_barcode}"
                )
                self.migration_report.add("Details", i18n.t("Duplicate loans (or failed twice)"))
                del self.failed[legacy_loan.item_barcode]
            return TransactionResult(False, False, "", "", "")

    def checkout_to_inactice_user(self, legacy_loan) -> TransactionResult:
        logging.info("Cannot check out to inactive user. Activating and trying again")
        user = self.get_user_by_barcode(legacy_loan.patron_barcode)
        expiration_date = user.get("expirationDate", datetime.isoformat(datetime.now()))
        user["expirationDate"] = datetime.isoformat(datetime.now() + timedelta(days=1))
        self.activate_user(user)
        logging.debug("Successfully Activated user")
        res = self.circulation_helper.check_out_by_barcode(legacy_loan)  # checkout_and_update
        self.migration_report.add("Details", res.migration_report_message)
        self.deactivate_user(user, expiration_date)
        logging.debug("Successfully Deactivated user again")
        self.migration_report.add("Details", i18n.t("Handled inactive users"))
        return res

    def handle_aged_to_lost_item(self, legacy_loan: LegacyLoan) -> TransactionResult:
        if self.circulation_helper.is_checked_out(legacy_loan):
            return TransactionResult(
                False,
                False,
                legacy_loan,
                i18n.t("Aged to lost and checked out"),
                i18n.t("Aged to lost and checked out"),
            )

        else:
            logging.debug("Setting Available")
            legacy_loan.next_item_status = "Available"
            self.set_item_status(legacy_loan)
            res_checkout = self.circulation_helper.check_out_by_barcode(legacy_loan)
            legacy_loan.next_item_status = "Aged to lost"
            self.set_item_status(legacy_loan)
            s = "Successfully Checked out Aged to lost item and put the status back"
            logging.info(s)
            self.migration_report.add("Details", s)
            return res_checkout

    def update_open_loan(self, folio_loan: dict, legacy_loan: LegacyLoan):
        due_date = du_parser.isoparse(str(legacy_loan.due_date))
        out_date = du_parser.isoparse(str(legacy_loan.out_date))
        renewal_count = legacy_loan.renewal_count
        try:
            loan_to_put = copy.deepcopy(folio_loan)
            del loan_to_put["metadata"]
            loan_to_put["dueDate"] = due_date.isoformat()
            loan_to_put["loanDate"] = out_date.isoformat()
            loan_to_put["renewalCount"] = renewal_count
            url = f"{self.folio_client.okapi_url}/circulation/loans/{loan_to_put['id']}"
            req = self.http_client.put(
                url,
                headers=self.folio_client.okapi_headers,
                json=loan_to_put,
            )
            if req.status_code == 422:
                error_message = json.loads(req.text)["errors"][0]["message"]
                s = f"Update open loan error: {error_message} {req.status_code}"
                self.migration_report.add("Details", s)
                logging.error(s)
                return False
            elif req.status_code in [201, 204]:
                self.migration_report.add(
                    "Details",
                    i18n.t("Successfully updated open loan") + f" ({req.status_code})",
                )
                return True
            else:
                self.migration_report.add(
                    "Details",
                    i18n.t("Update open loan error http status") + f": {req.status_code}",
                )
                req.raise_for_status()
            logging.debug("Updating open loan was successful")
            return True
        except HTTPError as exception:
            logging.error(
                f"{req.status_code} PUT FAILED Extend loan to {loan_to_put['dueDate']}"
                f"\t {url}\t{json.dumps(loan_to_put)}"
            )
            traceback.print_exc()
            logging.error(exception)
            return False

    def handle_previously_failed_loans(self, loan):
        if loan["item_id"] in self.failed:
            s = "Loan succeeded but failed previously. Removing from failed    "
            logging.info(s)
            del self.failed[loan["item_id"]]

    def declare_lost(self, folio_loan):
        declare_lost_url = f"/circulation/loans/{folio_loan['id']}/declare-item-lost"
        logging.debug(f"Declare lost url:{declare_lost_url}")
        due_date = du_parser.isoparse(folio_loan["dueDate"])
        data = {
            "declaredLostDateTime": datetime.isoformat(due_date + timedelta(days=1)),
            "comment": "Created at migration. Date is due date + 1 day",
            "servicePointId": str(self.task_configuration.fallback_service_point_id),
        }
        logging.debug(f"Declare lost data: {json.dumps(data, indent=4)}")
        if self.folio_put_post(declare_lost_url, data, "POST", i18n.t("Declare item as lost")):
            self.migration_report.add("Details", i18n.t("Successfully declared loan as lost"))
        else:
            logging.error(f"Unsuccessfully declared loan {folio_loan} as lost")
            self.migration_report.add("Details", i18n.t("Unsuccessfully declared loan as lost"))

    def claim_returned(self, folio_loan):
        claim_returned_url = f"/circulation/loans/{folio_loan['id']}/claim-item-returned"
        logging.debug(f"Claim returned url:{claim_returned_url}")
        due_date = du_parser.isoparse(folio_loan["dueDate"])
        data = {
            "itemClaimedReturnedDateTime": datetime.isoformat(due_date + timedelta(days=1)),
            "comment": "Created at migration. Date is due date + 1 day",
        }
        logging.debug(f"Claim returned data:\t{json.dumps(data)}")
        if self.folio_put_post(claim_returned_url, data, "POST", i18n.t("Declare item as lost")):
            self.migration_report.add(
                "Details", i18n.t("Successfully declared loan as Claimed returned")
            )
        else:
            logging.error(f"Unsuccessfully declared loan {folio_loan} as Claimed returned")
            self.migration_report.add(
                "Details",
                i18n.t(
                    "Unsuccessfully declared loan %{loan} as Claimed returned", loan=folio_loan
                ),
            )

    def set_item_status(self, legacy_loan: LegacyLoan):
        try:
            # Get Item by barcode, update status.
            item_path = f'item-storage/items?query=(barcode=="{legacy_loan.item_barcode}")'
            item_url = f"{self.folio_client.okapi_url}/{item_path}"
            resp = self.http_client.get(item_url, headers=self.folio_client.okapi_headers)
            resp.raise_for_status()
            data = resp.json()
            folio_item = data["items"][0]
            folio_item["status"]["name"] = legacy_loan.next_item_status
            if self.update_item(folio_item):
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "Successfully set item status to %{status}",
                        status=legacy_loan.next_item_status,
                    ),
                )
                logging.debug(
                    f"Successfully set item with barcode "
                    f"{legacy_loan.item_barcode} to {legacy_loan.next_item_status}"
                )
            else:
                if legacy_loan.item_barcode not in self.failed:
                    self.failed[legacy_loan.item_barcode] = legacy_loan
                logging.error(
                    f"Error when setting item with barcode "
                    f"{legacy_loan.item_barcode} to {legacy_loan.next_item_status}"
                )
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "Error setting item status to %{status}",
                        status=legacy_loan.next_item_status,
                    ),
                )
        except Exception as ee:
            logging.error(
                f"{resp.status_code} when trying to set item with barcode "
                f"{legacy_loan.item_barcode} to {legacy_loan.next_item_status} {ee}"
            )
            raise ee

    def activate_user(self, user):
        user["active"] = True
        self.update_user(user)
        self.migration_report.add("Details", i18n.t("Successfully activated user"))

    def deactivate_user(self, user, expiration_date):
        user["expirationDate"] = expiration_date
        user["active"] = False
        self.update_user(user)
        self.migration_report.add("Details", i18n.t("Successfully deactivated user"))

    def update_item(self, item):
        url = f'/item-storage/items/{item["id"]}'
        return self.folio_put_post(url, item, "PUT", i18n.t("Update item"))

    def update_user(self, user):
        url = f'/users/{user["id"]}'
        self.folio_put_post(url, user, "PUT", i18n.t("Update user"))

    def get_user_by_barcode(self, barcode):
        url = f'{self.folio_client.okapi_url}/users?query=(barcode=="{barcode}")'
        resp = self.http_client.get(url, headers=self.folio_client.okapi_headers)
        resp.raise_for_status()
        data = resp.json()
        return data["users"][0]

    def folio_put_post(self, url, data_dict, verb, action_description=""):
        full_url = f"{self.folio_client.okapi_url}{url}"
        try:
            if verb == "PUT":
                resp = self.http_client.put(
                    full_url,
                    headers=self.folio_client.okapi_headers,
                    json=data_dict,
                )
            elif verb == "POST":
                resp = self.http_client.post(
                    full_url,
                    headers=self.folio_client.okapi_headers,
                    json=data_dict,
                )
            else:
                raise Exception("Bad verb")
            if resp.status_code == 422:
                error_message = json.loads(resp.text)["errors"][0]["message"]
                logging.error(error_message)
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "%{action} error: %{message}",
                        action=action_description,
                        message=error_message,
                    ),
                )
                resp.raise_for_status()
            elif resp.status_code in [201, 204]:
                self.migration_report.add(
                    "Details",
                    i18n.t("Successfully %{action}", action=action_description)
                    + f" ({resp.status_code})",
                )
            else:
                self.migration_report.add(
                    "Details",
                    i18n.t(
                        "%{action} error. http status: %{status}",
                        action=action_description,
                        status=resp.status_code,
                    ),
                )

                resp.raise_for_status()
            return True
        except HTTPError as exception:
            logging.error(f"{resp.status_code}. {verb} FAILED for {url}")
            traceback.print_exc()
            logging.info(exception)
            return False

    def change_due_date(self, folio_loan, legacy_loan):
        try:
            api_path = f"{folio_loan['id']}/change-due-date"
            api_url = f"{self.folio_client.okapi_url}/circulation/loans/{api_path}"
            body = {"dueDate": du_parser.isoparse(str(legacy_loan.due_date)).isoformat()}
            req = self.http_client.post(
                api_url, headers=self.folio_client.okapi_headers, json=body
            )
            if req.status_code == 422:
                error_message = json.loads(req.text)["errors"][0]["message"]
                self.migration_report.add(
                    "Details", i18n.t("Change due date error") + f": {error_message}"
                )
                logging.info(
                    f"{error_message}\t",
                )
                self.migration_report.add("Details", error_message)
                return False
            elif req.status_code == 201:
                self.migration_report.add(
                    "Details", i18n.t("Successfully changed due date") + f" ({req.status_code})"
                )
                return True, json.loads(req.text), None
            elif req.status_code == 204:
                self.migration_report.add(
                    "Details", i18n.t("Successfully changed due date") + f" ({req.status_code})"
                )
                return True, None, None
            else:
                self.migration_report.add(
                    "Details",
                    i18n.t("Update open loan error http status"),
                    f": {req.status_code}",
                )
                req.raise_for_status()
        except HTTPError as exception:
            logging.info(
                f"{req.status_code} POST FAILED Change Due Date to {api_url}\t{json.dumps(body)})"
            )
            traceback.print_exc()
            logging.info(exception)
            return False, None, None


def timings(t0, t0func, num_objects):
    avg = num_objects / (time.time() - t0)
    elapsed = time.time() - t0
    elapsed_func = time.time() - t0func
    return (
        f"Total objects: {num_objects}\tTotal elapsed: {elapsed:.2f}\t"
        f"Average per object: {avg:.2f}\tElapsed this time: {elapsed_func:.2f}"
    )


def print_smtp_warning():
    s = """
          _____   __  __   _____   ______   ___
         / ____| |  \/  | |_   _| |  __  | |__ \\
        | (___   | \  / |   | |   | |__|_|    ) |
         \___ \  | |\/| |   | |   | |        / /
        |_____/  |_|  |_|   |_|   |_|       (_)
    """  # noqa: E501, W605
    print(s)
